#!/usr/local/bin/perl4 
#
# Copyright (C) 1996-1998 by the Board of Trustees
#    of Leland Stanford Junior University.
# 
# This file is part of the SimOS distribution. 
# See LICENSE file for terms of the license. 
#

# Mark A. Heinrich
# Stanford University
# FlashLite/Mipsy Statistics Post-Processor
# September 21, 1993

if (($#ARGV+1) < 2) {
   die "Usage: $0 [-h] <stat file> <MIPSY log file>\n";
}
while (1) {
   $next_option = shift(@ARGV);
   if ($next_option eq "-h") {
      $HISTMODE = 1;
   }
   else {
      unshift(@ARGV, $next_option);
      last;
   }
}
$APPL = shift(@ARGV);
$STATFILE = "$APPL.stats";
$MIPSYLOG = shift(@ARGV);
$OUTPUT = $MIPSYLOG;

# Get some old FL stats that are now in the MIPSYLOG file
open(MIPSYFILE, $MIPSYLOG) || die "Can't open $MIPSYLOG: $!\n";
while (<MIPSYFILE>) {

#
# Get Number of Processors
#
   ?^CONFIG\s+NumCPUs\s+(\d+).*$? && ($numProcs = $1);

#
# Number of Instructions and Thread runtimes
#
   if (/^M_EVENT_C(\d+)\s+Insts\s+(\d+).*$/) {
      $curProc = $1;
      $instruction_count += $2;
      $numInstructions[$1] = $2;
   }

   if (/^M_EVENT_C(\d+)\s+FP-Insts\s+(\d+).*$/) {
      $instruction_count += $2;
      $numInstructions[$1] += $2;
   }

   if (/^M_EVENT_C(\d+)\s+CPI\s+(\d+.\d+).*$/) {
      $CPI[$1] = $2;
   }

#
# Loads, Stores, Cache Stats
#
   if (/^C(\d+)\s+U\s+D\s+Read\s+Misses:\s+(\d+)\s+Reads:\s+(\d+).*$/) {
      $processor_loads += $3;
      $processorLoadArray[$1] = $3;
   }
   if (/^C(\d+)\s+U\s+D\s+Write\s+Misses:\s+(\d+)\s+Writes:\s+(\d+).*$/) {
      $processor_stores += $3;
      $processorStoreArray[$1] = $3;
   }

   # Include Icache misses in Get references (loads)
   if (/^C(\d+)\s+U\s+S\s+IMisses\s+(\d+)\s+IRefs:\s+(\d+).*$/) {
      $processor_load_misses += $2;
      $processorLdMissArray[$1] += $2;
   }

   if (/^C(\d+)\s+U\s+S\s+DMisses:\s+GetMisses\s+(\d+)\s+GetRefs\s+(\d+).*$/) {
      $processor_load_misses += $2;
      $processorLdMissArray[$1] += $2;
   }

   if (/^C(\d+)\s+U\s+S\s+DMisses:\s+GetXMisses\s+(\d+)\s+GetXRefs\s+(\d+).*$/) {
      $processor_store_misses += $2;
      $processorStMissArray[$1] += $2;
   }

   if (/^C(\d+)\s+U\s+S\s+UpgMisses:\s+(\d+)\s+UpgRefs:\s+(\d+).*$/) {
      $processor_stores += $3;
      $processor_upgrade_misses += $2;
      $processorStMissArray[$1] += $2;
   }

   if (/^M_EVENT_C(\d+)\s+Prefetches\s+(\d+).*$/) {
      $processor_prefs += $2;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchL1Hits\s+(\d+).*$/) {
      $processor_pref_hits += $2;
      $processor_pref_misses = $processor_prefs - $processor_pref_hits -
	 $processor_dropped_prefs - $processor_fullMHT_pref_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchL2Hits\s+(\d+).*$/) {
      $processor_pref_hits += $2;
      $processor_pref_misses = $processor_prefs - $processor_pref_hits -
	 $processor_dropped_prefs - $processor_fullMHT_pref_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchMerges\s+(\d+).*$/) {
      $processor_dropped_prefs += $2;
      $processor_pref_misses = $processor_prefs - $processor_pref_hits -
	 $processor_dropped_prefs - $processor_fullMHT_pref_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchMHTStall\s+(\d+).*$/) {
      $processor_fullMHT_pref_misses += $2;
      $processor_pref_misses = $processor_prefs - $processor_pref_hits -
	 $processor_dropped_prefs - $processor_fullMHT_pref_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchExclusives\s+(\d+).*$/) {
      $processor_prefxs += $2;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchXL1Hits\s+(\d+).*$/) {
      $processor_prefx_hits += $2;
      $processor_prefx_misses = $processor_prefxs - $processor_prefx_hits -
	 $processor_dropped_prefxs - $processor_fullMHT_prefx_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchXL2Hits\s+(\d+).*$/) {
      $processor_prefx_hits += $2;
      $processor_prefx_misses = $processor_prefxs - $processor_prefx_hits -
	 $processor_dropped_prefxs - $processor_fullMHT_prefx_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchXMerges\s+(\d+).*$/) {
      $processor_dropped_prefxs += $2;
      $processor_prefx_misses = $processor_prefxs - $processor_prefx_hits -
	 $processor_dropped_prefxs - $processor_fullMHT_prefx_misses;
   }

   if (/^M_EVENT_C(\d+)\s+PrefetchXMHTStall\s+(\d+).*$/) {
      $processor_fullMHT_prefx_misses += $2;
      $processor_prefx_misses = $processor_prefxs - $processor_prefx_hits -
	 $processor_dropped_prefxs - $processor_fullMHT_prefx_misses;
   }

   if (/^M_EVENT_C(\d+)\s+SyncOpStallTime\s+(\d+).*$/) {
      $syncWaitTimeArray[$1] = $2;
   }


#    if (/^U I\$ Get Retries\s+(\d+),\s+D\$ Get Retries\s+(\d+),\s+D\$ GetXRetries\s+(\d+).*$/) {
#       $IcacheGetRetries += $1;
#       $DcacheGetRetries += $2;
#       $DcacheGetRetryArray[$curProc] = $2;
#       $DcacheGetXRetries += $3;
#       # Make sure this makes sense
#       &assert(($loadCheck - $firstLevelLoadMisses[$curProc]) == $2);
#       &assert(($storeCheck - $firstLevelStoreMisses[$curProc]) == $3);

#       # Adjust number of store misses (save load for later)
#       $processor_store_misses -= $3;
#       $processorStMissArray[$curProc] -= $3;
#    }

#
# Synchronization
#
   if (/^M_EVENT_C(\d+)\s+LLStallTime\s+(\d+).*$/) {
      $LLStallTime[$1] = $2;
   }
   if (/^M_EVENT_C(\d+)\s+SCStallTime\s+(\d+).*$/) {
      $SCStallTime[$1] = $2;
   }

   if (/^SOLO_C(\d+)\s+Barriers\s+(\d+)\s+BarrierStall\s+(\d+).*$/) {
      $barWaitTimeArray[$1] = $3;
   }
   if (/^SOLO_C(\d+)\s+Locks\s+(\d+)\s+LockStall\s+(\d+).*$/) {
      $lockWaitTimeArray[$1] += $3;
   }
   if (/^SOLO_C(\d+)\s+Unlocks\s+(\d+)\s+UnlockStall\s+(\d+).*$/) {
      $lockWaitTimeArray[$1] += $3;
   }
   if (/^SOLO_C(\d+)\s+WaitFors\s+(\d+)\s+WaitForStall\s+(\d+)$/) {
      $waitForWaitTimeArray[$1] = $3;
   }


# # Not supported in Mipsy
#    if (/^processor_spin_lock_wait_time\s+(\d+).*$/) {
#       $spinLockWaitTimeArray[$numProcs-1] = $1;
#    }
#    if (/^processor_delay_wait_time\s+(\d+).*$/) {
#       $delayWaitTimeArray[$numProcs-1] = $1;
#    }
#    if (/^processor_getsub_wait_time\s+(\d+).*$/) {
#       $getsubWaitTimeArray[$numProcs-1] = $1;
#    }
#    if (/^processor_pause_wait_time\s+(\d+).*$/) {
#       $pauseWaitTimeArray[$numProcs-1] = $1;
#    }
}
close(MIPSYFILE);

# Save only the FlashLite statistics from the output file
open(OUTFILE, $OUTPUT) || die "Can't open $OUTPUT: $!\n";
while (<OUTFILE>) {
   /^total_cycles\s+(\d+).*$/ && ($total_cycles += $1);
   /^total_instructions\s+(\d+).*$/ && ($total_instructions += $1);
   /^total_ALU_instructions\s+(\d+).*$/ && ($total_ALU_instructions += $1);
   /^total_special_ALU_instructions\s+(\d+).*$/ && ($total_special_ALU_instructions += $1);
   /^total_branch_instructions\s+(\d+).*$/ && ($total_branch_instructions += $1);
   /^total_special_branch_instructions\s+(\d+).*$/ && ($total_special_branch_instructions += $1);
   /^total_ldst_instructions\s+(\d+).*$/ && ($total_ldst_instructions += $1);
   /^total_interface_instructions\s+(\d+).*$/ && ($total_interface_instructions += $1);
   if (/^total_(([PN]I|SW|IO)[^_]+)_latency_cycles\s+(\d+).*$/) {
      $latencies{$1} = $3;
   }
   if (/^total_(([PN]I|SW|IO)[^_]+)_occupancy_cycles\s+(\d+).*$/) {
      $occupancies{$1} = $3;
   }
   next unless /^BEGIN_PROC (\d+)/;
   $curProc = 0;
   @FL_OUTPUT = <OUTFILE>;
   last;
}

# Parse the FlashLite Statistics
foreach (@FL_OUTPUT) {
   if (/^BEGIN_PROC\s+(\d+).*$/) {
      $curProc = $1;
   }
   ?^simulation_time\s+(\d+).*$? && ($simulation_time = $1);
   ?^latency_factor\s+(\d+).*$? && ($latencyFactor = $1);

   if (/^total_execution_time\s+(\d+).*$/) {
      $executionTime = $1;
   }
   if (/^thread_execution_time_since_reset\s+(\d+).*$/) {
      $executionTimeSinceReset[$curProc] = $1;
   }

   if (/^messages_processed\s+(\d+)/) {
      if ($messages_processed != $totalCount) {
	 printf("WARNING: totalCount (%d) != messages_processed (%d) on node %d\n", $totalCount, $messages_processed, $curProc);
      }
      $messages_processed += $1;
   }
   if (/^(([PN]I|SW|IO)[^_]\w+)\s+(\d+).*$/ ) {
      $count{$1} += $3;
      $totalCount += $3;
   }
   if (/^Greatest_buffer_num_used\s+(\d+).*$/) {
      if ($1 > $greatest_buffer) {
	 $greatest_buffer = $1;
      }
      $greatest_buffer_sum += $1;
   }
   if (/^PI_wait_count\s+(\d+).*$/) {
      $piBufferWaitCount = $1;
   }
   if (/^PI_wait_time\s+(\d+).*$/) {
      if ($piBufferWaitCount > 0) {
	 $piBufferWaitLatency = $1/$piBufferWaitCount;
      }
      else {
	 $piBufferWaitLatency = 0;
      }
      if ($piBufferWaitLatency > $piBufferWaitLatency_max) {
	 $piBufferWaitLatency_max = $piBufferWaitLatency;
      }
      $piBufferWaitLatency_sum += $piBufferWaitLatency;
   }
   if (/^NI_req_wait_count\s+(\d+).*$/) {
      $niReqBufferWaitCount = $1;
   }
   if (/^NI_req_wait_time\s+(\d+).*$/) {
      if ($niReqBufferWaitCount > 0) {
	 $niReqBufferWaitLatency = $1/$niReqBufferWaitCount;
      }
      else {
	 $niReqBufferWaitLatency = 0;
      }
      if ($niReqBufferWaitLatency > $niReqBufferWaitLatency_max) {
	 $niReqBufferWaitLatency_max = $niReqBufferWaitLatency;
      }
      $niReqBufferWaitLatency_sum += $niReqBufferWaitLatency;
   }
   if (/^NI_reply_wait_count\s+(\d+).*$/) {
      $niRepBufferWaitCount = $1;
   }
   if (/^NI_reply_wait_time\s+(\d+).*$/) {
      if ($niRepBufferWaitCount > 0) {
	 $niRepBufferWaitLatency = $1/$niRepBufferWaitCount;
      }
      else {
	 $niReqBufferWaitLatency = 0;
      }
      if ($niRepBufferWaitLatency > $niRepBufferWaitLatency_max) {
	 $niRepBufferWaitLatency_max = $niRepBufferWaitLatency;
      }
      $niRepBufferWaitLatency_sum += $niRepBufferWaitLatency;
   }

   /^PI_memory_ops\s+(\d+).*$/ && ($pi_memory_ops += $1);
   /^NI_memory_ops\s+(\d+).*$/ && ($ni_memory_ops += $1);
   /^DP_memory_ops\s+(\d+).*$/ && ($dp_memory_ops += $1);
   /^dc_cache_misses\s+(\d+).*$/ && ($ddcache_misses += $1);
   /^dc_icache_misses\s+(\d+).*$/ && ($dicache_misses += $1);
   /^dc_cache_writebacks\s+(\d+).*$/ && ($ddcache_writebacks += $1);
   if (/^memory_ops\s+(\d+).*$/) {
      $total_memory_ops_sum += $1;
      if ($1 != 0) {
	 $memLatency = $memTime/$1;
      }
      else {
	 $memLatency = 0;
      }
      if ($memLatency > $memLatency_max) {
	 $memLatency_max = $memLatency;
      }
      $memLatency_sum += $memLatency;      
   }
   /^memory_reads\s+(\d+).*$/ && ($memory_reads += $1);
   /^memory_writes\s+(\d+).*$/ && ($memory_writes += $1);
   /^directory_memory_reads\s+(\d+).*$/ && ($dir_memory_reads += $1);
   /^directory_memory_writes\s+(\d+).*$/ && ($dir_memory_writes += $1);
   /^useless_memory_reads\s+(\d+).*$/ && ($useless_reads += $1);

#
# Pre-SysADBus Stall Time
#
   if (/^processor_write_stall_time\s+(\d+).*$/) {
      $writeStallTimeArray[$curProc] = $1;
   }
   if (/^processor_num_naks\s+(\d+).*$/) {
      $numNaks += $1;
   }

#
# SysADBus Stall Time
#
   if (/^read_SysAD_stall_time\s+(\d+).*$/) {
      $readSysADStallTime += $1;
   }
   if (/^write_SysAD_stall_time\s+(\d+).*$/) {
      $writeSysADStallTime += $1;
   }
   if (/^reply_SysAD_stall_time\s+(\d+).*$/) {
      $replySysADStallTime += $1;
   }
   if (/^total_SysAD_read_requests\s+(\d+).*$/) {
      $totalSysADReadRequests += $1;
   }
   if (/^total_SysAD_write_requests\s+(\d+).*$/) {
      $totalSysADWriteRequests += $1;
   }
   if (/^total_SysAD_reply_requests\s+(\d+).*$/) {
      $totalSysADReplyRequests += $1;
   }

#
# Misses
#
   if (/^processor_local_read_misses\s+(\d+).*$/) {
      $localReadMisses += $1;
      $localReadMissArray[$curProc] = $1;
   }
   if (/^processor_rem_read_misses\s+(\d+).*$/) {
      $remReadMisses += $1;
      $remReadMissArray[$curProc] = $1;
   }
   if (/^processor_local_write_misses\s+(\d+).*$/) {
      $localWriteMisses += $1;
#      $localWriteMissArray[$curProc] = $1;
   }
   if (/^processor_rem_write_misses\s+(\d+).*$/) {
      $remWriteMisses += $1;
#      $remWriteMissArray[$curProc] = $1;
   }
   if (/^processor_local_read_miss_latency\s+(\d+).*$/) {
      $localReadMissLatency += $1;
      $localReadMissLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_read_miss_latency\s+(\d+).*$/) {
      $remReadMissLatency += $1;
      $remReadMissLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_read_hist\[(\d+)\]\s+(\d+).*$/) {
      $localReadHist[$1] += $2;
      if ($localReadHist[$1] > $readHistMax) {
	 $readHistMax = $localReadHist[$1];
      }
   } 
   if (/^processor_rem_read_hist\[(\d+)\]\s+(\d+).*$/) {
      $remReadHist[$1] += $2;
      if ($remReadHist[$1] > $readHistMax) {
	 $readHistMax = $remReadHist[$1];
      }
   } 
   if (/^processor_local_write_miss_latency\s+(\d+).*$/) {
      $localWriteMissLatency += $1;
      $localWriteMissLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_write_miss_latency\s+(\d+).*$/) {
      $remWriteMissLatency += $1;
      $remWriteMissLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_write_hist\[(\d+)\]\s+(\d+).*$/) {
      $localWriteHist[$1] += $2;
      if ($localWriteHist[$1] > $writeHistMax) {
	 $writeHistMax = $localWriteHist[$1];
      }
   } 
   if (/^processor_rem_write_hist\[(\d+)\]\s+(\d+).*$/) {
      $remWriteHist[$1] += $2;
      if ($remWriteHist[$1] > $writeHistMax) {
	 $writeHistMax = $remWriteHist[$1];
      }
   } 

#
# Merges
#
   if (/^processor_local_read_merges\s+(\d+).*$/) {
      $localReadMerges += $1;
      $processorLdMergeArray[$curProc] += $1;
   }
   if (/^processor_rem_read_merges\s+(\d+).*$/) {
      $remReadMerges += $1;
      $processorLdMergeArray[$curProc] += $1;
   }
   if (/^processor_local_write_merges\s+(\d+).*$/) {
      $localWriteMerges += $1;
      $processorStMergeArray[$curProc] += $1;
   }
   if (/^processor_rem_write_merges\s+(\d+).*$/) {
      $remWriteMerges += $1;
      $processorStMergeArray[$curProc] += $1;
   }
   if (/^processor_local_read_merge_latency\s+(\d+).*$/) {
      $localReadMergeLatency += $1;
      $localReadMergeLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_read_merge_latency\s+(\d+).*$/) {
      $remReadMergeLatency += $1;
      $remReadMergeLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_read_merge_hist\[(\d+)\]\s+(\d+).*$/) {
      $localReadMergeHist[$1] += $2;
      if ($localReadMergeHist[$1] > $readMergeHistMax) {
	 $readMergeHistMax = $localReadMergeHist[$1];
      }
   } 
   if (/^processor_rem_read_merge_hist\[(\d+)\]\s+(\d+).*$/) {
      $remReadMergeHist[$1] += $2;
      if ($remReadMergeHist[$1] > $readMergeHistMax) {
	 $readMergeHistMax = $remReadMergeHist[$1];
      }
   } 
   if (/^processor_local_write_merge_latency\s+(\d+).*$/) {
      $localWriteMergeLatency += $1;
      $localWriteMergeLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_write_merge_latency\s+(\d+).*$/) {
      $remWriteMergeLatency += $1;
      $remWriteMergeLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_write_merge_hist\[(\d+)\]\s+(\d+).*$/) {
      $localWriteMergeHist[$1] += $2;
      if ($localWriteMergeHist[$1] > $writeMergeHistMax) {
	 $writeMergeHistMax = $localWriteMergeHist[$1];
      }
   } 
   if (/^processor_rem_write_merge_hist\[(\d+)\]\s+(\d+).*$/) {
      $remWriteMergeHist[$1] += $2;
      if ($remWriteMergeHist[$1] > $writeMergeHistMax) {
	 $writeMergeHistMax = $remWriteMergeHist[$1];
      }
   } 
#
# Conflicts
#
   if (/^processor_local_read_conflicts\s+(\d+).*$/) {
      $localReadConflicts += $1;
      $localReadConflictArray[$curProc] = $1;
      $processorLdConflictArray[$curProc] += $1;
   }
   if (/^processor_rem_read_conflicts\s+(\d+).*$/) {
      $remReadConflicts += $1;
      $remReadConflictArray[$curProc] = $1;
      $processorLdConflictArray[$curProc] += $1;
   }
   if (/^processor_local_write_conflicts\s+(\d+).*$/) {
      $localWriteConflicts += $1;
      $processorStConflictArray[$curProc] += $1;
   }
   if (/^processor_rem_write_conflicts\s+(\d+).*$/) {
      $remWriteConflicts += $1;
      $processorStConflictArray[$curProc] += $1;
   }
   if (/^processor_local_read_conflict_latency\s+(\d+).*$/) {
      $localReadConflictLatency += $1;
      $localReadConflictLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_read_conflict_latency\s+(\d+).*$/) {
      $remReadConflictLatency += $1;
      $remReadConflictLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_read_conflict_hist\[(\d+)\]\s+(\d+).*$/) {
      $localReadConflictHist[$1] += $2;
      if ($localReadConflictHist[$1] > $readConflictHistMax) {
	 $readConflictHistMax = $localReadConflictHist[$1];
      }
   } 
   if (/^processor_rem_read_conflict_hist\[(\d+)\]\s+(\d+).*$/) {
      $remReadConflictHist[$1] += $2;
      if ($remReadConflictHist[$1] > $readConflictHistMax) {
	 $readConflictHistMax = $remReadConflictHist[$1];
      }
   } 
   if (/^processor_local_write_conflict_latency\s+(\d+).*$/) {
      $localWriteConflictLatency += $1;
#      $localWriteConflictLatencyArray[$curProc] = $1;
   }
   if (/^processor_rem_write_conflict_latency\s+(\d+).*$/) {
      $remWriteConflictLatency += $1;
#      $remWriteConflictLatencyArray[$curProc] = $1;
   }
   if (/^processor_local_write_conflict_hist\[(\d+)\]\s+(\d+).*$/) {
      $localWriteConflictHist[$1] += $2;
      if ($localWriteConflictHist[$1] > $writeConflictHistMax) {
	 $writeConflictHistMax = $localWriteConflictHist[$1];
      }
   } 
   if (/^processor_rem_write_conflict_hist\[(\d+)\]\s+(\d+).*$/) {
      $remWriteConflictHist[$1] += $2;
      if ($remWriteConflictHist[$1] > $writeConflictHistMax) {
	 $writeConflictHistMax = $remWriteConflictHist[$1];
      }
   } 


#
# Invalidation Histograms
#
   if (/^inval_hist\[(\d+)\]\s+(\d+).*$/) {
      $invalHist{$curProc, $1} = $2;
   } 

#
# Queue Histograms
#
   if (/^pi_input_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $piInputQueueHist{$curProc, $1} = $2;
   }
   if (/^ni_input_req_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $niInputReqQueueHist{$curProc, $1} = $2;
   }
   if (/^ni_input_rep_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $niInputRepQueueHist{$curProc, $1} = $2;
   }
   if (/^ni_output_req_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $niOutputReqQueueHist{$curProc, $1} = $2;
   }
   if (/^ni_output_rep_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $niOutputRepQueueHist{$curProc, $1} = $2;
   }
   if (/^req_network_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $reqNetworkQueueHist{$curProc, $1} = $2;
   }
   if (/^rep_network_queue_hist\[(\d+)\]\s+(\d+).*$/) {
      $repNetworkQueueHist{$curProc, $1} = $2;
   }

   ?^dc_cache_size\s+(\d+).*$? && ($ddcache_size = $1);
   ?^dc_assoc\s+(\d+).*$? && ($ddcache_assoc = $1);
   ?^dc_replacement_policy\s+(\w+).*$? && ($ddcache_replacement = $1);
   /^dc_cache_ops\s+(\d+).*$/ && ($ddcache_ops += $1);
   /^dc_cache_reads\s+(\d+).*$/ && ($ddcache_reads += $1);
   /^dc_cache_writes\s+(\d+).*$/ && ($ddcache_writes += $1);
   /^dc_cache_hits\s+(\d+).*$/ && ($ddcache_hits += $1);
   /^dc_cache_read_hits\s+(\d+).*$/ && ($ddcache_read_hits += $1);
   /^dc_cache_write_hits\s+(\d+).*$/ && ($ddcache_write_hits += $1);
   /^dc_cache_read_misses\s+(\d+).*$/ && ($ddcache_read_misses += $1);
   /^dc_cache_write_misses\s+(\d+).*$/ && ($ddcache_write_misses += $1);

   ?^dc_icache_size\s+(\d+).*$? && ($dicache_size = $1);
   ?^dc_icache_assoc\s+(\d+).*$? && ($dicache_assoc = $1);
   ?^dc_icache_replacement_policy\s+(\w+).*$? && ($dicache_replacement = $1);
   /^dc_icache_ops\s+(\d+).*$/ && ($dicache_ops += $1);
   /^dc_icache_reads\s+(\d+).*$/ && ($dicache_reads += $1);
   /^dc_icache_writes\s+(\d+).*$/ && ($dicache_writes += $1);
   /^dc_icache_hits\s+(\d+).*$/ && ($dicache_hits += $1);
   /^dc_icache_read_hits\s+(\d+).*$/ && ($dicache_read_hits += $1);
   /^dc_icache_write_hits\s+(\d+).*$/ && ($dicache_write_hits += $1);
   /^dc_icache_read_misses\s+(\d+).*$/ && ($dicache_read_misses += $1);
   /^dc_icache_write_misses\s+(\d+).*$/ && ($dicache_write_misses += $1);

   /^SW_queue_final_count\s+(\d+).*$/ && ($swqueue_final += $1);
   /^SW_queue_re-entries\s+(\d+).*$/ && ($swqueue_retries += $1);
   /^reads_cancelled_by_full_SW_queue\s+(\d+).*$/ && ($swqueue_cancels += $1);
   if (/^PI_queue_max_count\s+(\d+).*$/) {
      $piInputQueue_max_sum += $1;
      if ($1 > $piInputQueue_max) {
	 $piInputQueue_max = $1;
      }
   }
   if (/^NI_req_in_queue_max_count\s+(\d+).*$/) {
      $niInputReqQueue_max_sum += $1;
      if ($1 > $niInputReqQueue_max) {
	 $niInputReqQueue_max = $1;
      }
   }
   if (/^NI_rep_in_queue_max_count\s+(\d+).*$/) {
      $niInputRepQueue_max_sum += $1;
      if ($1 > $niInputRepQueue_max) {
	 $niInputRepQueue_max = $1;
      }
   }
   if (/^Req_out_queue_max_count\s+(\d+).*$/) {
      $niOutputReqQueue_max_sum += $1;
      if ($1 > $niOutputReqQueue_max) {
	 $niOutputReqQueue_max = $1;
      }
   }
   if (/^Rep_out_queue_max_count\s+(\d+).*$/) {
      $niOutputRepQueue_max_sum += $1;
      if ($1 > $niOutputRepQueue_max) {
	 $niOutputRepQueue_max = $1;
      }
   }
#   /^time_in_memory_queue\s+(\d+).*$/ && ($memTime = $1);
   if (/^SysAD_occupancy\s+(\d+).*$/) {
      $sysADBusOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($sysADBusOccupancy > $sysADBusOccupancy_max) {
	 $sysADBusOccupancy_max = $sysADBusOccupancy;
      }
      $sysADBusOccupancy_sum += $sysADBusOccupancy;
   }
   if (/^Inbox_occupancy\s+(\d+).*$/) {
      $inboxOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($inboxOccupancy > $inboxOccupancy_max) {
	 $inboxOccupancy_max = $inboxOccupancy;
      }
      $inboxOccupancy_sum += $inboxOccupancy;
   }
   if (/^DP_occupancy\s+(\d+).*$/) {
      $dpOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($dpOccupancy > $dpOccupancy_max) {
	 $dpOccupancy_max = $dpOccupancy;
      }
      $dpOccupancy_sum += $dpOccupancy;
   }
   if (/^time_in_memory\s+(\d+).*$/) {
      $memTime = $1;
      $memOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($memOccupancy > $memOccupancy_max) {
	 $memOccupancy_max = $memOccupancy;
      }
      $memOccupancy_sum += $memOccupancy;
   }
   if (/^NI_request_occupancy\s+(\d+).*$/) {
      $niReqOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($niReqOccupancy > $niReqOccupancy_max) {
	 $niReqOccupancy_max = $niReqOccupancy;
      }
      $niReqOccupancy_sum += $niReqOccupancy;
   }
   if (/^NI_reply_occupancy\s+(\d+).*$/) {
      $niRepOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($niRepOccupancy > $niRepOccupancy_max) {
	 $niRepOccupancy_max = $niRepOccupancy;
      }
      $niRepOccupancy_sum += $niRepOccupancy;
   }
   if (/^NI_request_in_occupancy\s+(\d+).*$/) {
      $niReqInOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($niReqInOccupancy > $niReqInOccupancy_max) {
	 $niReqInOccupancy_max = $niReqInOccupancy;
      }
      $niReqInOccupancy_sum += $niReqInOccupancy;
   }
   if (/^NI_reply_in_occupancy\s+(\d+).*$/) {
      $niRepInOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($niRepInOccupancy > $niRepInOccupancy_max) {
	 $niRepInOccupancy_max = $niRepInOccupancy;
      }
      $niRepInOccupancy_sum += $niRepInOccupancy;
   }
   if (/^PI_in_occupancy\s+(\d+).*$/) {
      $piInOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($piInOccupancy > $piInOccupancy_max) {
	 $piInOccupancy_max = $piInOccupancy;
      }
      $piInOccupancy_sum += $piInOccupancy;
   }
   if (/^PI_out_occupancy\s+(\d+).*$/) {
      $piOutOccupancy = (($1/$executionTimeSinceReset[$curProc])*100.0);
      if ($piOutOccupancy > $piOutOccupancy_max) {
	 $piOutOccupancy_max = $piOutOccupancy;
      }
      $piOutOccupancy_sum += $piOutOccupancy;
   }
   if (/^Inbox_addr_match\s+(\d+).*$/) {
      $inboxAddrMatch += $1;
   }
   if (/^Inbox_addr_match_delay_count\s+(\d+).*$/) {
      $inboxAddrMatchDelayCount += $1;
   }
   if (/^Inbox_addr_match_delay_time\s+(\d+).*$/) {
      $inboxAddrMatchDelayTime += $1;
   }
   if (/^Inbox_DP_stall_time\s+(\d+).*$/) {
      $inboxDPStallTime += $1;
   }
   if (/^Inbox_DP_stall_count\s+(\d+).*$/) {
      $inboxDPStallCount += $1;
   }
   if (/^DP_Mem_stall_time\s+(\d+).*$/) {
      $dpMemStallTime += $1;
   }
   if (/^DP_Outbox_stall_time\s+(\d+).*$/) {
      $dpOutboxStallTime += $1;
   }
   if (/^Outbox_PI_stall_time\s+(\d+).*$/) {
      $outboxPIStallTime += $1;
   }
   if (/^Outbox_NI_stall_time\s+(\d+).*$/) {
      $outboxNIStallTime += $1;
   }
   if (/^PI_SysAD_stall_time\s+(\d+).*$/) {
      $piSysADStallTime += $1;
   }
   if (/^total_SysAD_PI_requests\s+(\d+).*$/) {
      $totalSysADPIRequests += $1;
   }
   if (/^processor_Cache_stall_time\s+(\d+).*$/) {
      $processorCacheStallTime += $1;
      $processorCacheStallTimeArray[$curProc] = $1;
   }
   if (/^read_invalidate_retries\s+(\d+).*$/) {
      $readInvalidateRetries += $1;
   }
   if (/^upgrade_invalidate_retries\s+(\d+).*$/) {
      $upgradeInvalidateRetries += $1;
   }
   if (/^processor_read_bus_id_wait_time\s+(\d+).*$/) {
      $readBusIDWaitTime += $1;
   }
   if (/^processor_write_bus_id_wait_time\s+(\d+).*$/) {
      $writeBusIDWaitTime += $1;
   }
   if (/^dirty_local_get_count\s+(\d+).*$/) {
      $dirtyLocalGetCount += $1;
   }
   if (/^dirty_local_getx_count\s+(\d+).*$/) {
      $dirtyLocalGetXCount += $1;
   }
}

# Open the final, nicely formatted statistics file

open(STAT, "> $STATFILE") || die "Can't open $STATFILE: $!\n";

&DO_FLASHLITE_PERF_STATS();
&PRINT_DIVIDER();
&DO_PROCESSOR_STATS();
&PRINT_DIVIDER();
&DO_SCACHE_STATS();
&PRINT_DIVIDER();
&DO_DPSIM_STATS();
&PRINT_DIVIDER();
&DO_HANDLER_STATS();
&PRINT_DIVIDER();
&DO_READ_CLASSIFICATION_STATS();
&PRINT_DIVIDER();
&DO_SW_QUEUE_STATS();
&PRINT_DIVIDER();
&DO_DDCACHE_STATS();
&PRINT_DIVIDER();
&DO_DICACHE_STATS();
&PRINT_DIVIDER();
&DO_MEMORY_STATS();
&PRINT_DIVIDER();
&DO_BUFFER_STATS(); 
&PRINT_DIVIDER();
&DO_QUEUE_STATS();
&PRINT_DIVIDER();
&DO_OCCUPANCY_STATS();
&PRINT_DIVIDER();
&DO_MACROPIPELINE_STALL_STATS();
close(STAT);
#system("raw.pl $OUTPUT") &&
#    die "raw.pl exited abnormally: $!\n";
exit(0);

sub numerically { $a <=> $b;}
sub PRINT_DIVIDER {
   printf(STAT "\n************************************************************\n\n");
}

sub DO_FLASHLITE_PERF_STATS {
   printf(STAT "Simulation Time: %d Hours %d Minutes %d Seconds\n",
	  $simulation_time/3600, ($simulation_time/60)%60, 
	  $simulation_time % 60);
   printf(STAT "FLASH Instructions Simulated: %.0lf\n", 
	  $instruction_count);
   printf(STAT "FlashLite Speed: %8.3lf (Instructions/Second)\n",
	  $instruction_count/$simulation_time);
   printf(STAT "Execution Time of Simulated Application: %.0lf cycles",
	  $executionTime/$latencyFactor);
}

sub DO_SCACHE_STATS {
   printf(STAT "Secondary Cache Statistics\n\n");
   printf(STAT "Miss Rate = %.2lf%%\n\n", (($processor_load_misses +
	$processor_store_misses + $processor_upgrade_misses + 
	$processor_pref_misses + $processor_prefx_misses + 
        $processor_fullMHT_pref_misses + $processor_fullMHT_prefx_misses)/
       ($processor_loads + 
	$processor_stores + $processor_prefs + $processor_prefxs))
	* 100.0);
   printf(STAT "%-25s %9s %12s\n", Operation, Count, Percentage); 
   printf(STAT "------------------------------------------------\n");
   printf(STAT "%-25s %9d %9.1lf\n", Load, $processor_loads, 
	($processor_loads/($processor_loads + $processor_stores +
	$processor_prefs + $processor_prefxs)) * 100.0);
   printf(STAT "\n%25s %9d %9.1lf\n", Hits, $processor_loads - 
	$processor_load_misses - ($localReadMerges + $remReadMerges), 
	(($processor_loads - $processor_load_misses - ($localReadMerges +
        $remReadMerges))/$processor_loads) * 100.0);
   printf(STAT "%25s %9d %9.1lf\n", Misses, $processor_load_misses, 
	($processor_load_misses/$processor_loads) * 100.0);
   printf(STAT "%25s %9d %9.1lf\n\n", Merges, ($localReadMerges + 
          $remReadMerges), (($localReadMerges + $remReadMerges)/
	  $processor_loads) * 100.0);
   if ($processor_load_misses != 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Local Home", 
	     $localReadMisses,
	     ($localReadMisses/($localReadMisses + $remReadMisses))*100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Remote Home", 
	     $remReadMisses,
	     ($remReadMisses/($localReadMisses + $remReadMisses))*100.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Remote Home", 0, 0.0);
   }
   if ($localReadMerges + $remReadMerges == 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Remote Home", 0, 0.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Local Home", 
	     $localReadMerges, ($localReadMerges/($localReadMerges + 
	     $remReadMerges))*100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Remote Home", 
	     $remReadMerges, ($remReadMerges/($localReadMerges + 
	     $remReadMerges))*100.0);
   }
   if ($localReadConflicts + $remReadConflicts == 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Remote Home", 0, 0.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Local Home", 
	  $localReadConflicts,
	  ($localReadConflicts/($localReadConflicts + $remReadConflicts))
	  *100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Remote Home", 
	  $remReadConflicts,
	  ($remReadConflicts/($localReadConflicts + $remReadConflicts))*100.0);
   }
   printf(STAT "\n%-25s %9d %9.1lf\n", Store, $processor_stores, 
	($processor_stores/($processor_loads + $processor_stores +
	$processor_prefs + $processor_prefxs)) * 100.0);
   printf(STAT "\n%25s %9d %9.1lf\n", Hits, $processor_stores - 
	$processor_store_misses - $processor_upgrade_misses - 
        ($localWriteMerges + $remWriteMerges), 
	(($processor_stores - $processor_store_misses - ($localWriteMerges
        + $remWriteMerges) -
	$processor_upgrade_misses)/$processor_stores) * 100.0);
   printf(STAT "%25s %9d %9.1lf\n", "Store Misses", $processor_store_misses, 
	($processor_store_misses/$processor_stores) * 100.0);
   printf(STAT "%25s %9d %9.1lf\n", "Upgrade Misses", 
	$processor_upgrade_misses, ($processor_upgrade_misses/
	$processor_stores) * 100.0);
   printf(STAT "%25s %9d %9.1lf\n\n", Merges, 
	  ($localWriteMerges + $remWriteMerges),
	  (($localWriteMerges + $remWriteMerges)/$processor_stores) * 100.0);
   if ($localWriteMisses + $remWriteMisses == 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Remote Home", 0, 0.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Local Home", $localWriteMisses,
	  ($localWriteMisses/($localWriteMisses + $remWriteMisses))*100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Misses w/ Remote Home", $remWriteMisses,
	  ($remWriteMisses/($localWriteMisses + $remWriteMisses))*100.0);
   }

   if ($localWriteMerges + $remWriteMerges == 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Remote Home", 0, 0.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Local Home", 
	     $localWriteMerges, ($localWriteMerges/($localWriteMerges + 
	     $remWriteMerges))*100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Merges w/ Remote Home", 
	     $remWriteMerges, ($remWriteMerges/($localWriteMerges + 
	     $remWriteMerges))*100.0);
   }
   if ($localWriteConflicts + $remWriteConflicts == 0) {
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Local Home", 0, 0.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Remote Home", 0, 0.0);
   }
   else {
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Local Home", 
	  $localWriteConflicts,
	  ($localWriteConflicts/($localWriteConflicts + $remWriteConflicts))*
	  100.0);
      printf(STAT "%-25s %9d %9.1lf\n", "Conflicts w/ Remote Home", 
	  $remWriteConflicts,
	  ($remWriteConflicts/($localWriteConflicts + $remWriteConflicts))*
	  100.0);
   }

   if ($processor_prefs > 0) {
      printf(STAT "\n%-25s %9d %9.1lf\n", Prefetch, $processor_prefs, 
	 ($processor_prefs/($processor_loads + $processor_stores +
	 $processor_prefs + $processor_prefxs)) * 100.0);
      printf(STAT "\n%25s %9d %9.1lf\n", Hits, $processor_prefs - 
	 $processor_pref_misses - $processor_dropped_prefs - 
         $processor_fullMHT_pref_misses, 
         (($processor_prefs - $processor_pref_misses - 
         $processor_dropped_prefs - $processor_fullMHT_pref_misses)/
	 $processor_prefs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", Misses, $processor_pref_misses, 
	 ($processor_pref_misses/$processor_prefs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", Dropped, $processor_dropped_prefs,
	 ($processor_dropped_prefs/$processor_prefs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", "Full MHT", 
	  $processor_fullMHT_pref_misses,
	 ($processor_fullMHT_pref_misses/$processor_prefs) * 100.0);
   }
   else {
      printf(STAT "\n%-25s %9d %9.1lf\n", Prefetch, 0, 0.0);
   }

   if ($processor_prefxs > 0) {
      printf(STAT "\n%-25s %9d %9.1lf\n", "Prefetch Exclusive", 
	 $processor_prefxs, ($processor_prefxs/($processor_prefxs + 
	 $processor_stores + $processor_prefs + $processor_prefxs)) * 100.0);
      printf(STAT "\n%25s %9d %9.1lf\n", Hits, $processor_prefxs - 
	 $processor_prefx_misses - $processor_dropped_prefxs -
         $processor_fullMHT_prefx_misses, (($processor_prefxs - 
	 $processor_prefx_misses - $processor_dropped_prefxs -
         $processor_fullMHT_prefx_misses)/$processor_prefxs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", Misses, $processor_prefx_misses, 
	 ($processor_prefx_misses/$processor_prefxs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", Dropped, $processor_dropped_prefxs,
	 ($processor_dropped_prefxs/$processor_prefxs) * 100.0);
      printf(STAT "%25s %9d %9.1lf\n", "Full MHT", 
	  $processor_fullMHT_prefx_misses,
	 ($processor_fullMHT_prefx_misses/$processor_prefxs) * 100.0);
   }
   else {
      printf(STAT "\n%-25s %9d %9.1lf\n", "Prefetch Exclusive", 0, 0.0);
   }

}

sub DO_DPSIM_STATS {
   printf(STAT "DPsim statistics\n\n");

   if ($total_branch_instructions > 0) {
   printf(STAT "%-40s %8d\n", "Total cycles @ 2 slots/cycle:", $total_cycles);
   printf(STAT "%-40s %8d\n", "Total instructions:", $total_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage slots filled:",
          $total_instructions/(2.0*$total_cycles));
   printf(STAT "\n%-40s %8d\n", "Total ALU instructions:",
          $total_ALU_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage ALU of total:",
          $total_ALU_instructions/$total_instructions);
   printf(STAT "\n%-40s %8d\n", "Total special ALU instructions:",
          $total_special_ALU_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage special ALU of ALU:",
          $total_special_ALU_instructions/$total_ALU_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage special ALU of total:",
          $total_special_ALU_instructions/$total_instructions);
   printf(STAT "\n%-40s %8d\n", "Total branch/jump instructions:",
          $total_branch_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage branch of total:",
          $total_branch_instructions/$total_instructions);
   printf(STAT "\n%-40s %8d\n", "Total special branch instructions:",
          $total_special_branch_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage special branch of branch:",
          $total_special_branch_instructions/$total_branch_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage special branch of total:",
          $total_special_branch_instructions/$total_instructions);
   printf(STAT "\n%-40s %8d\n", "Total ld/st instructions:",
          $total_ldst_instructions);
   printf(STAT "%-40s %5.2f\n", "Percentage ld/st of total:",
          $total_ldst_instructions/$total_instructions);
   printf(STAT "\n%-40s %8d\n", "Total interface instructions:",
          $total_interface_instructions);
   printf(STAT "%-40s %5.2f\n\n", "Percentage interface of total:",
          $total_interface_instructions/$total_instructions);
   }
   else {
      printf(STAT "DPsim not used for this run\n\n");
   }
}

sub DO_READ_CLASSIFICATION_STATS {
   if ($processor_load_misses != 0) {
      printf(STAT "Breakdown of Read Distributions:\n\n");
      printf(STAT "%-40s %8d (%5.2lf%%)\n", "Local, satisfied locally", 
	  $piLocalGetCount - $niLocalPutCount, (($piLocalGetCount - 
	  $niLocalPutCount)/$processor_load_misses)*100.0);
      printf(STAT "%-40s %8d (%5.2lf%%)\n", "Local, satisfied remote", 
	  $niLocalPutCount, ($niLocalPutCount/$processor_load_misses)*100.0);
      printf(STAT "%-40s %8d (%5.2lf%%)\n", "Remote, clean at home",
	  $niRemotePutCount - $dirtyLocalGetCount - $niSharingWritebackCount,
	  (($niRemotePutCount - $dirtyLocalGetCount - $niSharingWritebackCount)
	   / $processor_load_misses) * 100.0);
      printf(STAT "%-40s %8d (%5.2lf%%)\n", "Remote, dirty at home",
	  $dirtyLocalGetCount, ($dirtyLocalGetCount/$processor_load_misses)*
	  100.0);
      printf(STAT "%-40s %8d (%5.2lf%%)\n", "Remote, dirty remote",
	  $niSharingWritebackCount, ($niSharingWritebackCount/
          $processor_load_misses) * 100.0);				     
   }
}

sub DO_HANDLER_STATS {
   printf(STAT "Handler Statistics\n\n");

   printf(STAT "%-40s %8d\n", "Read-Invalidate retries:", 
	  $readInvalidateRetries);
   printf(STAT "%-40s %8d\n", "Upgrade-Invalidate retries:", 
	  $upgradeInvalidateRetries);
   printf(STAT "%-40s %8d\n", "Total Number of Naks:", $numNaks);
   foreach $i (30..$#localReadHist) {
      $numLongReqs += $localReadHist[$i] + $remReadHist[$i] + 
	 $localWriteHist[$i] + $remWriteHist[$i] + $localReadMergeHist[$i]
	 + $remReadMergeHist[$i] + $localWriteMergeHist[$i] + 
         $remWriteMergeHist[$i] + $localReadConflictHist[$i] +
	 $remReadConflictHist[$i] + $localWriteConflictHist[$i] +
	 $remWriteConflictHist[$i];
   }
   printf(STAT "%-40s %8d\n", "Total Request Latencies above 300:", 
	  $numLongReqs);
   printf(STAT "%-40s %8d\n", "Dirty Local Get Count:", $dirtyLocalGetCount);
   printf(STAT "%-40s %8d\n", "Dirty Local GetX Count:", $dirtyLocalGetXCount);
   if ($messages_processed == 0) {
      return;
   }

   $hack = 100;
   foreach $handler (keys(%count)) {
      $count{$handler} = $count{$handler} . $hack++;
      $lines{$count{$handler}} = $handler . ',' . 0;
   }
   printf(STAT "\n%-28s %5s %7s %14s %7s %10s\n", Handler, Cycles, Count, 
	  "Frequency (%)",Path,Occupancy);
   printf(STAT "-----------------------------------------------------------------------------\n");
   foreach $counts (reverse sort numerically keys(%lines)) {
      ($handler, $cycleCount) = split(/,/, $lines{$counts});
      chop $counts;
      chop $counts;
      chop $counts;

      # Insert tests for certain handlers.  This info used in the
      # read breakdown section
      if ($handler eq "PILocalGet") {
	 $piLocalGetCount = $counts;
      }
      if ($handler eq "NILocalPut") {
	 $niLocalPutCount = $counts;
      }
      if ($handler eq "NIRemotePut") {
	 $niRemotePutCount = $counts;
      }
      if ($handler eq "NISharingWriteback") {
	 $niSharingWritebackCount = $counts;
      }
      $path = $latencies{$handler};
      if ($path) {
         if ($counts != 0) {
            $path /= $counts;
         }
      }
      else {
         $path = "n/a";
      }
      $occupancy = $occupancies{$handler};
      if ($occupancy) {
         if ($counts != 0) {
            $occupancy /= $counts;
         }
      }
      else {
         $occupancy = "n/a";
      }
      if ($path eq "n/a") {
         printf(STAT "%-30s %5d %8d %8.1lf %11s %8s\n", $handler, $cycleCount, $counts,
	   ($counts/$messages_processed)*100.0,$path,$occupancy);
      }
      else {
         printf(STAT "%-30s %5d %8d %8.1lf %11.1lf %8.1lf\n", $handler, $cycleCount, $counts,
	   ($counts/$messages_processed)*100.0,$path,$occupancy);
      }
   }
   printf(STAT "\nTotals %38d %8.1lf\n", $totalCount, 100.0);
   if ($totalCount != $messages_processed) {
      print STAT "WARNING: TotalCount != messages_processed ($messages_processed)\n";
   }
   printf(STAT "\nHandler Calls Per Miss = %5.2lf\n", $totalCount/
	($processor_load_misses + $processor_store_misses +
         $processor_upgrade_misses + 
         $processor_pref_misses + $processor_prefx_misses));
}

sub DO_DDCACHE_STATS {
   if ($ddcache_ops == 0) {
      return;
   }
   printf(STAT "Directory Data Cache Statistics\n\n");
   printf(STAT "DD\$ Size = %d KB\n", $ddcache_size >> 10);
   printf(STAT "DD\$ Associativity = %d\n", $ddcache_assoc);
   printf(STAT "DD\$ Replacement Policy = $ddcache_replacement\n\n");
   printf(STAT "Hit Rate  = %5.1lf%%\n", ($ddcache_hits/$ddcache_ops) *
      100.0);
   printf(STAT "Miss Rate = %5.1lf%%\n\n", ($ddcache_misses/$ddcache_ops) *
      100.0);
   if (($ddcache_hits + $ddcache_misses) != $ddcache_ops) {
      printf(STAT "WARNING: Hits + Misses != Total Ops\n");
   }
   printf(STAT "%-18s %8s %12s\n", Operation, Count, Percentage); 
   printf(STAT "-----------------------------------------\n");
   printf(STAT "%-18s %8d %9.1lf\n", Read, $ddcache_reads, ($ddcache_reads/
	$ddcache_ops) * 100.0);
   printf(STAT "\n%18s %8d %9.1lf\n", Hits, $ddcache_read_hits, 
	($ddcache_read_hits/$ddcache_reads) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", Misses, $ddcache_read_misses, 
      ($ddcache_read_misses/$ddcache_reads) * 100.0);

   if ($ddcache_read_hits + $ddcache_read_misses != $ddcache_reads) {
      printf(STAT "WARNING: Read Hits + Read Misses != Total Reads\n");
   }

   printf(STAT "\n%-18s %8d %9.1lf\n", Write, $ddcache_writes, 
	($ddcache_writes/$ddcache_ops) * 100.0);
   printf(STAT "\n%18s %8d %9.1lf\n", Hits, $ddcache_write_hits,
      ($ddcache_write_hits/$ddcache_writes) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", Misses, $ddcache_write_misses,
      ($ddcache_write_misses/$ddcache_writes) * 100.0);

   if ($ddcache_write_hits + $ddcache_write_misses != $ddcache_writes) {
      printf(STAT "WARNING: Write Hits + Write Misses != Total Writes\n");
   }
}

sub DO_DICACHE_STATS {
   if ($dicache_ops == 0) {
      return;
   }
   printf(STAT "Directory Instruction Cache Statistics\n\n");
   printf(STAT "DI\$ Size = %d KB\n", $dicache_size >> 10);
   printf(STAT "DI\$ Associativity = %d\n", $dicache_assoc);
   printf(STAT "DI\$ Replacement Policy = $dicache_replacement\n\n");
   printf(STAT "Hit Rate  = %5.1lf%%\n", ($dicache_hits/$dicache_ops) *
      100.0);
   printf(STAT "Miss Rate = %5.1lf%%\n\n", ($dicache_misses/$dicache_ops) *
      100.0);
   if (($dicache_hits + $dicache_misses) != $dicache_ops) {
      printf(STAT "WARNING: Hits + Misses != Total Ops\n");
   }
   printf(STAT "%-18s %8s %12s\n", Operation, Count, Percentage); 
   printf(STAT "-----------------------------------------\n");
   printf(STAT "%-18s %8d %9.1lf\n", Read, $dicache_reads, ($dicache_reads/
	$dicache_ops) * 100.0);
   printf(STAT "\n%18s %8d %9.1lf\n", Hits, $dicache_read_hits, 
	($dicache_read_hits/$dicache_reads) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", Misses, $dicache_read_misses, 
      ($dicache_read_misses/$dicache_reads) * 100.0);

   if ($dicache_read_hits + $dicache_read_misses != $dicache_reads) {
      printf(STAT "WARNING: Read Hits + Read Misses != Total Reads\n");
   }

   printf(STAT "\n%-18s %8d %9.1lf\n", Write, $dicache_writes, 
	($dicache_writes/$dicache_ops) * 100.0);
}

sub DO_SW_QUEUE_STATS {
   print STAT "Software Queue Statistics\n\n";
   if ($swqueue_final != 0) {
      print STAT "WARNING: SW Queue final count ($swqueue_final) is not 0!\n";
   }
   printf(STAT "SW Queue requests returned to queue = %d\n", $swqueue_retries);
   printf(STAT "Requests NAKed because SW Queue count is full = %d\n",
      $swqueue_cancels);
   
}

sub DO_BUFFER_STATS {
   print STAT "Data Buffer Statistics\n\n";
   printf(STAT "%-31s %5d\n", "Highest Data Buffer Used:", $greatest_buffer);
   printf(STAT "%-31s %5d\n", "Avg Highest Buffer Used:", $greatest_buffer_sum/
	$numProcs);
   printf(STAT "%-31s %5d\n", "Avg PI Buffer Wait Latency:", 
	  $piBufferWaitLatency_sum/($numProcs*$latencyFactor));
   printf(STAT "%-31s %5d\n", "Max PI Buffer Wait Latency:", 
	  $piBufferWaitLatency_max/$latencyFactor);
   printf(STAT "%-31s %5d\n", "Avg NI Req Buffer Wait Latency:", 
	  $niReqBufferWaitLatency_sum/($numProcs * $latencyFactor));
   printf(STAT "%-31s %5d\n", "Max NI Req Buffer Wait Latency:", 
	  $niReqBufferWaitLatency_max/$latencyFactor);
   printf(STAT "%-31s %5d\n", "Avg NI Rep Buffer Wait Latency:", 
	  $niRepBufferWaitLatency_sum/($numProcs*$latencyFactor));
   printf(STAT "%-31s %5d\n", "Max NI Rep Buffer Wait Latency:", 
	  $niRepBufferWaitLatency_max/$latencyFactor);
}	

sub DO_QUEUE_STATS {
   print STAT "Queue Statistics\n\n";
   printf(STAT "Maximum PI Input Queue Length = %d, Avg of Max = %d\n", 
	  $piInputQueue_max, $piInputQueue_max_sum/$numProcs);
   printf(STAT "Maximum NI Input Request Queue Length = %d, Avg of Max = %d\n",
	  $niInputReqQueue_max, $niInputReqQueue_max_sum/$numProcs);
   printf(STAT "Maximum NI Input Reply Queue Length = %d, Avg of Max = %d\n",
      $niInputRepQueue_max, $niInputRepQueue_max_sum/$numProcs);
   printf(STAT "Maximum NI Output Request Queue Length = %d, Avg of Max = %d\n", $niOutputReqQueue_max, $niOutputReqQueue_max_sum/$numProcs);
   printf(STAT "Maximum NI Output Reply Queue Length = %d, Avg of Max = %d\n",
      $niOutputRepQueue_max, $niOutputRepQueue_max_sum/$numProcs);
   printf(STAT "\nPI Input Queue Histogram\n");
   printf(STAT "%2s %5s %4s %4s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
   printf(STAT "----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      foreach $j (0..0) {
	 printf(STAT "%5d ", $piInputQueueHist{$i, $j});
      }
      foreach $j (1..2) {
	 printf(STAT "%4d ", $piInputQueueHist{$i, $j});
      }
      foreach $j (3..15) {
	 printf(STAT "%3d ", $piInputQueueHist{$i, $j});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nNI Input Request Queue Histogram\n");
   printf(STAT "%2s %5s %4s %4s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
   printf(STAT "----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      foreach $j (0..0) {
	 printf(STAT "%5d ", $niInputReqQueueHist{$i, $j});
      }
      foreach $j (1..2) {
	 printf(STAT "%4d ", $niInputReqQueueHist{$i, $j});
      }
      foreach $j (3..15) {
	 printf(STAT "%3d ", $niInputReqQueueHist{$i, $j});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nNI Input Reply Queue Histogram\n");
   printf(STAT "%2s %5s %4s %4s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
   printf(STAT "----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      foreach $j (0..0) {
	 printf(STAT "%5d ", $niInputRepQueueHist{$i, $j});
      }
      foreach $j (1..2) {
	 printf(STAT "%4d ", $niInputRepQueueHist{$i, $j});
      }
      foreach $j (3..15) {
	 printf(STAT "%3d ", $niInputRepQueueHist{$i, $j});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nNI Output Request Queue Histogram\n");
   printf(STAT "%2s %5s %4s %4s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
   printf(STAT "----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      foreach $j (0..0) {
	 printf(STAT "%5d ", $niOutputReqQueueHist{$i, $j});
      }
      foreach $j (1..2) {
	 printf(STAT "%4d ", $niOutputReqQueueHist{$i, $j});
      }
      foreach $j (3..15) {
	 printf(STAT "%3d ", $niOutputReqQueueHist{$i, $j});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nNI Output Reply Queue Histogram\n");
   printf(STAT "%2s %5s %4s %4s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15);
   printf(STAT "----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      foreach $j (0..0) {
	 printf(STAT "%5d ", $niOutputRepQueueHist{$i, $j});
      }
      foreach $j (1..2) {
	 printf(STAT "%4d ", $niOutputRepQueueHist{$i, $j});
      }
      foreach $j (3..15) {
	 printf(STAT "%3d ", $niOutputRepQueueHist{$i, $j});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nRequest Network Queue Histogram\n");
   printf(STAT "%2s %4s %4s %4s %4s %4s %4s %4s %4s %4s %4s\n",
	  P, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20);
   printf(STAT "----------------------------------------------------------\n");
   for ($i=0; $i < $numProcs; $i++) {
      printf(STAT "p%x ", $i);
      for ($j=0; $j < 20; $j +=2) {
	 printf(STAT "%4d ", $reqNetworkQueueHist{$i, $j} +
		$reqNetworkQueueHist{$i, $j+1});
      }
      printf(STAT "\n");
   }
   printf(STAT "\nReply Network Queue Histogram\n");
   printf(STAT "%2s %4s %4s %4s %4s %4s %4s %4s %4s %4s %4s\n",
	  P, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20);
   printf(STAT "----------------------------------------------------------\n");
   for ($i=0; $i < $numProcs; $i++) {
      printf(STAT "p%x ", $i);
      for ($j=0; $j < 20; $j +=2) {
	 printf(STAT "%4d ", $repNetworkQueueHist{$i, $j} + 
		$repNetworkQueueHist{$i, $j+1});
      }
      printf(STAT "\n");
   }
}

sub DO_MEMORY_STATS {
   if ($total_memory_ops_sum == 0) {
      return;
   }
   printf(STAT "Memory Ops\n\n");
   printf(STAT "%-18s %8s %12s\n", Source, Count, Percentage); 
   printf(STAT "-----------------------------------------\n");
   printf(STAT "%-18s %8d %9.1lf\n", "Inbox from PI", $pi_memory_ops,
      ($pi_memory_ops/$total_memory_ops_sum) * 100.0);
   printf(STAT "%-18s %8d %9.1lf\n", "Inbox from NI", $ni_memory_ops,
      ($ni_memory_ops/$total_memory_ops_sum) * 100.0);
   printf(STAT "%-18s %8d %9.1lf\n", "DP Handler", $dp_memory_ops,
      ($dp_memory_ops/$total_memory_ops_sum) * 100.0);
   printf(STAT "%-18s %8d %9.1lf\n", "DD\$ Miss", $ddcache_misses,
      ($ddcache_misses/$total_memory_ops_sum) * 100.0);
   printf(STAT "%-18s %8d %9.1lf\n", "DD\$ Writeback", $ddcache_writebacks,
      ($ddcache_writebacks/$total_memory_ops_sum) * 100.0);
   printf(STAT "%-18s %8d %9.1lf\n", "DI\$ Miss", $dicache_misses,
      ($dicache_misses/$total_memory_ops_sum) * 100.0);
   printf(STAT "\nTotals %20d %9.1lf\n\n", $total_memory_ops_sum, 100.0);

   if (($pi_memory_ops + $ni_memory_ops + $dp_memory_ops + $ddcache_misses +
      $ddcache_writebacks + $dicache_misses) != $total_memory_ops_sum) {
      printf(STAT "\nWARNING: Memory Ops do not add to total\n");
   }
   printf(STAT "%-18s %8s %12s\n", Operation, Count, Percentage); 
   printf(STAT "-----------------------------------------\n");
   printf(STAT "%-18s %8d %9.1lf\n", "Memory Reads", $memory_reads,
      ($memory_reads/$total_memory_ops_sum) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", "Useless Reads", $useless_reads,
      ($useless_reads/$memory_reads) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", "Appl Reads", $memory_reads -
      $useless_reads - $dir_memory_reads, (($memory_reads - $useless_reads -
	 $dir_memory_reads)/$memory_reads) * 100.0);
   printf(STAT "%18s %8d %9.1lf\n", "Dir Reads", $dir_memory_reads,
      ($dir_memory_reads/$memory_reads) * 100.0);
   printf(STAT "\n%-18s %8d %9.1lf\n", "Memory Writes", $memory_writes,
      ($memory_writes/$total_memory_ops_sum) * 100.0);
   if ($memory_writes != 0) {
      printf(STAT "%18s %8d %9.1lf\n", "Appl Writes", $memory_writes -
	     $dir_memory_writes, (($memory_writes - $dir_memory_writes)/
	     $memory_writes) * 100.0);
      printf(STAT "%18s %8d %9.1lf\n", "Dir Writes", $dir_memory_writes,
	     ($dir_memory_writes/$memory_writes) * 100.0);
   }
   else {
      printf(STAT "%18s %8d %9.1lf\n", "Appl Writes", 0, 0, 0.0);
      printf(STAT "%18s %8d %9.1lf\n", "Dir Writes", 0, 0.0);
   }

   printf(STAT "Avg Memory Latency: %3d\n", $memLatency_sum/($numProcs *
							     $latencyFactor));
   printf(STAT "Max Memory Latency: %3d\n", $memLatency_max/$latencyFactor);
}

sub DO_PROCESSOR_STATS {
   print STAT "Processor Statistics\n\n";
   printf(STAT "\n%11s %8s %8s %8s %8s %8s %9s\n", "Busy", 'Pref', "Read", 
	  "Write", "Fence", "Synch", "Total");
   printf(STAT "-------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      $totalSynchronizationTime[$i] = $barWaitTimeArray[$i] + 
	 $lockWaitTimeArray[$i] + $spinLockWaitTimeArray[$i] +
	 $pauseWaitTimeArray[$i] + $delayWaitTimeArray[$i] + 
         $getsubWaitTimeArray[$i] + $waitForWaitTimeArray[$i];

      $busy = (($executionTimeSinceReset[$i] - 
	      ($processorCacheStallTimeArray[$i] +
	      $localReadMissLatencyArray[$i] + $remReadMissLatencyArray[$i] +
	      $localReadMergeLatencyArray[$i] + $remReadMergeLatencyArray[$i] +
	      $localReadConflictLatencyArray[$i] + 
	      $remReadConflictLatencyArray[$i] + $writeStallTimeArray[$i] +
	      $syncWaitTimeArray[$i] + $totalSynchronizationTime[$i]))/
              $executionTimeSinceReset[$i]) * 100.0;
      $busy_sum += $busy;
      printf(STAT "p%x %8.2lf %8.2lf %8.2lf %8.2lf %8.2lf %8.2lf %10.0lf\n", $i, 
	     $busy,
	     ($processorCacheStallTimeArray[$i]/$executionTimeSinceReset[$i])
	     * 100.0,
	     (($localReadMissLatencyArray[$i] + $remReadMissLatencyArray[$i] +
	       $localReadMergeLatencyArray[$i] + $remReadMergeLatencyArray[$i]
	       + $localReadConflictLatencyArray[$i] + 
	       $remReadConflictLatencyArray[$i])/$executionTimeSinceReset[$i])
	     * 100.0, 
	     ($writeStallTimeArray[$i]/$executionTimeSinceReset[$i]) * 100.0,
	     ($syncWaitTimeArray[$i]/
	     $executionTimeSinceReset[$i]) * 100.0, 
	     ($totalSynchronizationTime[$i]/
	     $executionTimeSinceReset[$i]) * 100.0, 
	     $executionTimeSinceReset[$i]/$latencyFactor);
   }
   printf(STAT "\n");
   printf(STAT "Avg. Processor Utilization: %4.1lf%%\n", $busy_sum/$numProcs);

   printf(STAT "\nPer Processor Invalidation Histogram\n\n");
   if ($numProcs <= 16) {
      printf(STAT "%2s %5s %5s %4s %3s %3s %3s %3s %3s %3s %3s %3s %2s %2s %2s %2s %2s %3s\n", P, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, "16+");
   }
   else {
      printf(STAT "%2s %5s %5d %4d %3d %3d %3d %3d %3d %3d %3d %3d %2d %2d %2d %2d %2d %3d+\n", P, 0, $numProcs/16, 2*$numProcs/16, 3*$numProcs/16, 4*$numProcs/16, 5*$numProcs/16, 6*$numProcs/16, 7*$numProcs/16, 8*$numProcs/16, 9*$numProcs/16, 10*$numProcs/16, 11*$numProcs/16, 12*$numProcs/16, 13*$numProcs/16, 14*$numProcs/16, 15*$numProcs/16, $numProcs);
   }

   printf(STAT "-----------------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x ", $i);
      $lastbin = 0;
      if ($numProcs <= 16) {
	 foreach $j (0..1) {
	    printf(STAT "%5d ", $invalHist{$i, $j});
	    $totalInvals += $invalHist{$i, $j} * $j;
	 }
	 foreach $j (2..2) {
	    printf(STAT "%4d ", $invalHist{$i, $j});
	    $totalInvals += $invalHist{$i, $j} * $j;
	 }
	 foreach $j (3..10) {
	    printf(STAT "%3d ", $invalHist{$i, $j});
	    $totalInvals += $invalHist{$i, $j} * $j;
	 }
	 foreach $j (11..19) {
	    if ($j < 16) {
	       printf(STAT "%2d ", $invalHist{$i, $j});
	       $totalInvals += $invalHist{$i, $j} * $j;
	    }
	    else {
	       $lastbin += $invalHist{$i, $j};
	    }
	 }
      }
      else {
	 foreach $j (0..1) {
	    $subtotal = 0;
	    $binStart = 4*$j;
	    $binEnd = ($numProcs/16) - 1 + (4*$j);
	    foreach $k ($binStart..$binEnd) {
	       $subtotal += $invalHist{$i, $k};
	       $totalInvals += $invalHist{$i, $k} * $k;
	    }
	    printf(STAT "%5d ", $subtotal);
	 }
	 foreach $j (2..2) {
	    $subtotal = 0;
	    $binStart = 4*$j;
	    $binEnd = ($numProcs/16) - 1 + (4*$j);
	    foreach $k ($binStart..$binEnd) {
	       $subtotal += $invalHist{$i, $k};
	       $totalInvals += $invalHist{$i, $k} * $k;
	    }
	    printf(STAT "%4d ", $subtotal);
	 }
	 foreach $j (3..10) {
	    $subtotal = 0;
	    $binStart = 4*$j;
	    $binEnd = ($numProcs/16) - 1 + (4*$j);
	    foreach $k ($binStart..$binEnd) {
	       $subtotal += $invalHist{$i, $k};
	       $totalInvals += $invalHist{$i, $k} * $k;
	    }
	    printf(STAT "%3d ", $subtotal);
	 }
	 foreach $j (11..19) {
	    $subtotal = 0;
	    $temp = ($numProcs/16) - 1 + (4*$j);
	    $binStart = 4*$j;
	    $binEnd = ($numProcs/16) - 1 + (4*$j);
	    if ($binEnd < $numProcs) {
	       foreach $k ($binStart..$binEnd) {
		  $subtotal += $invalHist{$i, $k};
		  $totalInvals += $invalHist{$i, $k} * $k;
	       }
	       printf(STAT "%2d ", $subtotal);
	    }
	    else {
	       foreach $k ($binStart..$binEnd) {
		  $lastbin += $invalHist{$i, $k};
		  $totalInvals += $invalHist{$i, $k} * $k;
	       }
	    }
	 }
      }
      printf(STAT "%2d ", $lastbin);
      printf(STAT "\n");
   }
   printf(STAT "Total Invalidations = %d\n", $totalInvals);   

   printf(STAT "\n%11s %8s %8s %8s %8s %8s %8s\n", "Barriers", "Locks", 
	  "Spin", "GetSub", "Delay", "Pause", "WaitFor");
   printf(STAT "-----------------------------------------------------------------\n");
   foreach $i (0..$numProcs-1) {
      if ($totalSynchronizationTime[$i] == 0) {
 	 next;
      }
      printf(STAT "p%x %8.2lf %8.2lf %8.2lf %8.2lf %8.2lf %8.2lf %8.2lf\n", $i,
	     ($barWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($lockWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($spinLockWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($getsubWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($delayWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($pauseWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0,
	     ($waitForWaitTimeArray[$i]/$totalSynchronizationTime[$i])*100.0);
   }

   printf(STAT "\nHere merges are included in all colums\n");
   printf(STAT "%14s %6s\n", "Insts", "CPI");
   printf(STAT "--------------------------\n");
   foreach $i (0..$numProcs-1) {
      printf(STAT "p%x %12.0lf %7.2lf\n", $i, $numInstructions[$i], $CPI[$i]);
   }
   printf(STAT "\n");

   if ($localReadMisses + $remReadMisses + $localReadConflicts + 
       $remReadConflicts != $processor_load_misses) {
      printf(STAT "WARNING: localReadMisses + remReadMisses != total\n");
      printf(STAT "Total misses = %d, local = %d, remote = %d, merges = %d\n",
	     $processor_load_misses, $localReadMisses, $remReadMisses,
	     $localReadMerges + $remReadMerges);
   }
   if ($localWriteMisses + $remWriteMisses + $localWriteConflicts + 
       $remWriteConflicts != 
       ($processor_store_misses + $processor_upgrade_misses)) {
      printf(STAT "WARNING: localWriteMisses + remWriteMisses != total\n");
      printf(STAT "Total misses = %d, local = %d, remote = %d, merges = %d\n",
	     $processor_store_misses + $processor_upgrade_misses, 
	     $localWriteMisses, $remWriteMisses, $localWriteMerges +
	     $remWriteMerges);
   }
   printf(STAT "Total Read Miss Stall Time = %d\n", ($localReadMissLatency +
	  $remReadMissLatency)/$latencyFactor);
   printf(STAT "Local Read Miss Stall Time = %d\n", $localReadMissLatency/
	  $latencyFactor);
   printf(STAT "Remote Read Miss Stall Time = %d\n", $remReadMissLatency/
	  $latencyFactor);
   if ($processor_load_misses != 0) {
      printf(STAT "\nAvg Read Miss Latency = %d\n", 
	  ($localReadMissLatency + $remReadMissLatency)/
	  ($processor_load_misses * $latencyFactor));
   }
   else {
      printf(STAT "Avg Read Miss Latency = 0\n");
   }

   if ($localReadMisses == 0) {
      printf(STAT "Avg Local ReadMissLatency = 0\n");
   }
   else {
      printf(STAT "Avg Local Read Miss Latency = %d\n",
	     $localReadMissLatency/($localReadMisses * $latencyFactor));
   }

   if ($remReadMisses == 0) {
      printf(STAT "Avg Remote Read Miss Latency = 0\n");
   }
   else {
      printf(STAT "Avg Remote Read Miss Latency = %d\n",
	     $remReadMissLatency/($remReadMisses * $latencyFactor));
   }

   if ($processor_load_misses != 0) {
      printf(STAT "%-25s %.2lf\n", "Avg Read Bus ID Wait Time:", 
	     $readBusIDWaitTime/($processor_load_misses * $latencyFactor));
   }
   else {
      printf(STAT "%-25s %.2lf\n", "Avg Read Bus ID Wait Time:", 0.0);
   }

   if (($processor_store_misses + $processor_upgrade_misses) > 0) {
      printf(STAT "%-25s %.2lf\n", "Avg Write Bus ID Wait Time:", 
	  $writeBusIDWaitTime/(($processor_store_misses + 
				$processor_upgrade_misses) * $latencyFactor));
   }
   else {
      printf(STAT "%-25s %.2lf\n", "Avg Write Bus ID Wait Time:", 0); 
   }

   printf(STAT "Avg SysADBus Occupancy = %4.1lf%%\n", $sysADBusOccupancy_sum/
	  $numProcs);
   printf(STAT "Max SysADBus Occupancy = %4.1lf%%\n", $sysADBusOccupancy_max);

   if ($totalSysADReadRequests != 0) {
      printf(STAT "\nAvg Read-SysADBus Stall Time = %.2lf\n",
	  $readSysADStallTime/($totalSysADReadRequests * $latencyFactor));
   }
   else {
      printf(STAT "\nAvg Read-SysADBus Stall Time = %.2lf\n", 0.0);
   }
   if ($totalSysADWriteRequests != 0) {
      printf(STAT "Avg Write-SysADBus Stall Time = %.2lf\n",
	  $writeSysADStallTime/($totalSysADWriteRequests * $latencyFactor));
   }
   else {
      printf(STAT "Avg Write-SysADBus Stall Time = %.2lf\n", 0.0);
   }
   if ($totalSysADReplyRequests != 0) {
      printf(STAT "\nAvg Reply-SysADBus Stall Time = %.2lf\n",
	  $replySysADStallTime/($totalSysADReplyRequests * $latencyFactor));
   }
   else {
      printf(STAT "\nAvg Reply-SysADBus Stall Time = %.2lf\n", 0.0);
   }
   if ($totalSysADPIRequests != 0) {
      printf(STAT "Avg PI-SysAD Stall Time = %.2lf\n", 
	  $piSysADStallTime/($totalSysADPIRequests * $latencyFactor));
   }
   else {
      printf(STAT "Avg PI-SysAD Stall Time = %.2lf\n", 0.0);
   }

   if ($HISTMODE) {
      printf(STAT "\nHistogram of Read misses\n");
      if ($readHistMax != 0) {
	 foreach $i (0 .. $#localReadHist) {
	    printf(STAT "Bin %3d %9d %-22s\n", $i * 10,
	     $localReadHist[$i] + $remReadHist[$i], "*"x((($localReadHist[$i]+
	     $remReadHist[$i])/$readHistMax)*30.0));
	 }
      }
   }
#   printf(STAT "\nHistogram of Remote Read misses\n");
#   foreach $i (0 .. $#remReadHist) {
#      printf(STAT "Bin %3d %9d\n", $i * 10, $remReadHist[$i]);
#   }
   
   if ($localReadMerges + $remReadMerges == 0) {
      printf(STAT "No Read Merges\n");
   }
   else {
      printf(STAT "Total Read Merge Stall Time = %d\n", 
	     ($localReadMergeLatency +
	      $remReadMergeLatency)/$latencyFactor);
      printf(STAT "Local Read Merge Stall Time = %d\n", $localReadMergeLatency/
	     $latencyFactor);
      printf(STAT "Remote Read Merge Stall Time = %d\n", $remReadMergeLatency/
	     $latencyFactor);
      printf(STAT "\nAvg Read Merge Latency = %d\n", 
	  ($localReadMergeLatency + $remReadMergeLatency)/
	  (($localReadMerges + $remReadMerges) * $latencyFactor));

      if ($localReadMerges == 0) {
	 printf(STAT "Avg Local ReadMergeLatency = 0\n");
      }
      else {
	 printf(STAT "Avg Local Read Merge Latency = %d\n",
	     $localReadMergeLatency/($localReadMerges * $latencyFactor));
      }

      if ($remReadMerges == 0) {
	 printf(STAT "Avg Remote Read Merge Latency = 0\n");
      }
      else {
	 printf(STAT "Avg Remote Read Merge Latency = %d\n",
		$remReadMergeLatency/($remReadMerges * $latencyFactor));
      }

      if ($HISTMODE) {
	 printf(STAT "\nHistogram of Read merges\n");
	 foreach $i (0 .. $#localReadMergeHist) {
	    printf(STAT "Bin %3d %9d %-22s\n", $i * 10,
	     $localReadMergeHist[$i] + $remReadMergeHist[$i], 
	     "*"x((($localReadMergeHist[$i]+
	     $remReadMergeHist[$i])/$readMergeHistMax)*30.0));
	 }
      }
   }

   if ($localReadConflicts + $remReadConflicts == 0) {
      printf(STAT "\nNo Read Conflicts\n");
   }
   else {
      printf(STAT "Total Read Conflict Stall Time = %d\n", 
	     ($localReadConflictLatency +
	      $remReadConflictLatency)/$latencyFactor);
      printf(STAT "Local Read Conflict Stall Time = %d\n",
	     $localReadConflictLatency/$latencyFactor);
      printf(STAT "Remote Read Conflict Stall Time = %d\n", 
	     $remReadConflictLatency/$latencyFactor);
      printf(STAT "\nAvg Read Conflict Latency = %d\n", 
	  ($localReadConflictLatency + $remReadConflictLatency)/
	  (($localReadConflicts + $remReadConflicts) * $latencyFactor));

      if ($localReadConflicts == 0) {
	 printf(STAT "Avg Local ReadConflictLatency = 0\n");
      }
      else {
	 printf(STAT "Avg Local Read Conflict Latency = %d\n",
	     $localReadConflictLatency/($localReadConflicts * $latencyFactor));
      }

      if ($remReadConflicts == 0) {
	 printf(STAT "Avg Remote Read Conflict Latency = 0\n");
      }
      else {
	 printf(STAT "Avg Remote Read Conflict Latency = %d\n",
		$remReadConflictLatency/($remReadConflicts * $latencyFactor));
      }

      if ($HISTMODE) {
	 printf(STAT "\nHistogram of Read conflicts\n");
	 foreach $i (0 .. $#localReadConflictHist) {
	    printf(STAT "Bin %3d %9d %-22s\n", $i * 10,
	     $localReadConflictHist[$i] + $remReadConflictHist[$i], 
	     "*"x((($localReadConflictHist[$i]+
	     $remReadConflictHist[$i])/$readConflictHistMax)*30.0));
	 }
      }
   }

   if (($processor_store_misses + $processor_upgrade_misses) > 0) {
      printf(STAT "\nAvg Write Miss Latency = %d\n", 
	  ($localWriteMissLatency + $remWriteMissLatency)/
	  (($processor_store_misses + $processor_upgrade_misses) * 
	   $latencyFactor));
   }
   else {
      printf(STAT "\nAvg Write Miss Latency = %d\n", 0);
   }

   if ($localWriteMisses == 0) {
      printf(STAT "Avg Local Write Miss Latency = 0\n");
   }
   else {
      printf(STAT "Avg Local Write Miss Latency = %d\n",
	     $localWriteMissLatency/($localWriteMisses * $latencyFactor));
   }

   if ($remWriteMisses == 0) {
      printf(STAT "Avg Remote Write Miss Latency = 0\n");
   }
   else {
      printf(STAT "Avg Remote Write Miss Latency = %d\n",
	     $remWriteMissLatency/($remWriteMisses * $latencyFactor));
   }

   if ($HISTMODE) {
      printf(STAT "\nHistogram of Write misses\n");
      if ($writeHistMax == 0) {
	 printf(STAT "No write miss latency\n");
      }
      else {
	 foreach $i (0 .. $#localWriteHist) {
	    printf(STAT "Bin %3d %9d %-22s\n", $i * 10, 
	     $localWriteHist[$i] + $remWriteHist[$i], "*"x(((
	     $localWriteHist[$i]+ $remWriteHist[$i])/$writeHistMax)*30));
	 }
      }
   }
#   printf(STAT "\nHistogram of Remote Write misses\n");
#   foreach $i (0 .. $#remWriteHist) {
#      printf(STAT "Bin %3d %9d\n", $i * 10,
#	     $remWriteHist[$i]);
#   }

   if ($localWriteMerges + $remWriteMerges != 0) {
      printf(STAT "\nAvg Write Merge Latency = %d\n", 
	  ($localWriteMergeLatency + $remWriteMergeLatency)/
	  (($localWriteMerges + $remWriteMerges) * $latencyFactor));
   }

   if ($localWriteMerges == 0) {
      printf(STAT "Avg Local Write Merge Latency = 0\n");
   }
   else {
      printf(STAT "Avg Local Write Merge Latency = %d\n",
	     $localWriteMergeLatency/($localWriteMerges * $latencyFactor));
   }

   if ($remWriteMerges == 0) {
      printf(STAT "Avg Remote Write Merge Latency = 0\n");
   }
   else {
      printf(STAT "Avg Remote Write Merge Latency = %d\n",
	     $remWriteMergeLatency/($remWriteMerges * $latencyFactor));
   }

   if ($writeMergeHistMax == 0) {
      printf(STAT "No write merge latency\n");
   }
   else {
      if ($HISTMODE) {
	 printf(STAT "\nHistogram of Write merges\n");
	 foreach $i (0 .. $#localWriteMergeHist) {
	    printf(STAT "Bin %3d %9d %-22s\n", $i * 10, 
	     $localWriteMergeHist[$i] + $remWriteMergeHist[$i], "*"x(((
	     $localWriteMergeHist[$i]+ $remWriteMergeHist[$i])/
             $writeMergeHistMax)*30));
	 }
      }
   }

   if ($localWriteConflicts + $remWriteConflicts == 0) {
      printf(STAT "\nNo write conflicts\n");
   }
   else {
      printf(STAT "\nAvg Write Conflict Latency = %d\n", 
	  ($localWriteConflictLatency + $remWriteConflictLatency)/
	  (($localWriteConflicts + $remWriteConflicts) * $latencyFactor));

      if ($localWriteConflicts == 0) {
	 printf(STAT "Avg Local Write Conflict Latency = 0\n");
      }
      else {
	 printf(STAT "Avg Local Write Conflict Latency = %d\n",
		$localWriteConflictLatency/($localWriteConflicts * 
					 $latencyFactor));
      }

      if ($remWriteConflicts == 0) {
	 printf(STAT "Avg Remote Write Conflict Latency = 0\n");
      }
      else {
	 printf(STAT "Avg Remote Write Conflict Latency = %d\n",
	     $remWriteConflictLatency/($remWriteConflicts * $latencyFactor));
      }

      if ($writeConflictHistMax == 0) {
	 printf(STAT "No write conflict latency\n");
      }
      else {
	 if ($HISTMODE) {
	    printf(STAT "\nHistogram of Write conflicts\n");
	    foreach $i (0 .. $#localWriteConflictHist) {
	       printf(STAT "Bin %3d %9d %-22s\n", $i * 10, 
	     $localWriteConflictHist[$i] + $remWriteConflictHist[$i], "*"x(((
	     $localWriteConflictHist[$i]+ $remWriteConflictHist[$i])/
             $writeConflictHistMax)*30));
	    }
	 }
      }
   }

   # --- Insert Mipsy stats here ---
   &PRINT_DIVIDER();
   close(STAT);
   system("mipsy_stats.pl $MIPSYLOG >> $STATFILE") && 
      die ("Inclusion of Mipsy stats failed\n");
   open (STAT, ">>$STATFILE") ||
      die ("Re-open of FL stats file failed\n");
   &PRINT_DIVIDER();
}

sub DO_OCCUPANCY_STATS {
   print STAT "Occupancy Statistics\n\n";

   printf(STAT "Avg Inbox Occupancy = %4.1lf%%\n", $inboxOccupancy_sum/
	  $numProcs);
   printf(STAT "Max Inbox Occupancy = %4.1lf%%\n", $inboxOccupancy_max);
   printf(STAT "Avg DP Occupancy = %4.1lf%%\n", $dpOccupancy_sum/$numProcs);
   printf(STAT "Max DP Occupancy = %4.1lf%%\n", $dpOccupancy_max);
   printf(STAT "Avg Memory Occupancy = %4.1lf%%\n", $memOccupancy_sum/
	  $numProcs);
   printf(STAT "Max Memory Occupancy = %4.1lf%%\n", $memOccupancy_max);
   printf(STAT "Avg PI In Occupancy = %4.1lf%%\n", 
	  $piInOccupancy_sum/$numProcs);
   printf(STAT "Max PI In Occupancy = %4.1lf%%\n", 
	  $piInOccupancy_max);
   printf(STAT "Avg PI Out Occupancy = %4.1lf%%\n", 
	  $piOutOccupancy_sum/$numProcs);
   printf(STAT "Max PI Out Occupancy = %4.1lf%%\n", 
	  $piOutOccupancy_max);
   printf(STAT "Avg NI Request In Occupancy = %4.1lf%%\n", 
	  $niReqInOccupancy_sum/$numProcs);
   printf(STAT "Max NI Request In Occupancy = %4.1lf%%\n", 
	  $niReqInOccupancy_max);
   printf(STAT "Avg NI Reply In Occupancy = %4.1lf%%\n", 
	  $niRepInOccupancy_sum/$numProcs);
   printf(STAT "Max NI Reply In Occupancy = %4.1lf%%\n", 
	  $niRepInOccupancy_max);
   printf(STAT "Avg NI Request Out Occupancy = %4.1lf%%\n", 
	  $niReqOccupancy_sum/$numProcs);
   printf(STAT "Max NI Request Out Occupancy = %4.1lf%%\n", 
	  $niReqOccupancy_max);
   printf(STAT "Avg NI Reply Out Occupancy = %4.1lf%%\n", 
	  $niRepOccupancy_sum/$numProcs);
   printf(STAT "Max NI Reply Out Occupancy = %4.1lf%%\n", 
	  $niRepOccupancy_max);
}

sub DO_MACROPIPELINE_STALL_STATS {
   print STAT "Macropipeline Stall Statistics\n\n";

   printf(STAT "Inbox-DP Address matches = %d\n", $inboxAddrMatch);
   printf(STAT "Inbox-DP Address matches delayed = %d\n", 
	  $inboxAddrMatchDelayCount);
   printf(STAT "Inbox-DP Address match delay time = %d\n", 
	  $inboxAddrMatchDelayTime/$latencyFactor);
   printf(STAT "Inbox-DP Stall Time = %d\n", $inboxDPStallTime/$latencyFactor);
   if ($inboxDPStallCount != 0) {
      printf(STAT "Inbox-DP Stall Count = %d (%.2lf)\n", $inboxDPStallCount,
	  ($inboxDPStallTime/$latencyFactor)/$inboxDPStallCount);
   }
   printf(STAT "DP-Memory Stall Time = %d\n", $dpMemStallTime/$latencyFactor);
   printf(STAT "DP-Outbox Stall Time = %d\n", $dpOutboxStallTime/
	  $latencyFactor);
   printf(STAT "Outbox-PI Stall Time = %d\n", $outboxPIStallTime/
	  $latencyFactor);
   printf(STAT "Outbox-NI Stall Time = %d\n", $outboxNIStallTime/
	  $latencyFactor);
   printf(STAT "PI-SysAD Stall Time = %d\n", $piSysADStallTime/$latencyFactor);
   printf(STAT "Processor-Cache Stall Time = %d\n", $processorCacheStallTime/
	  $latencyFactor);
}



