/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */
 
/*****************************************************************
 * 
 * $Author: nikos $
 * $Date
 *****************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <netinet/in.h>

#ifdef __alpha
#define _LANGUAGE_C
#define _MIPS_SZINT 32
/* #include "elf-sgi.h" */
#endif
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "file_formats.h"
#include "symbols.h"
#include "simutil.h"
#include "symfile.h"
#include "assoctab.h"
#include "tcl_init.h"
#include "sim_error.h"

#define MIPS_MDEBUG             ".mdebug"

#define ADDR_IN_RANGE(_addr, _lo, _hi)  (((uint)(_addr) >= (uint)(_lo)) && ((uint)(_addr) < (uint)(_hi)))


/*
 *   types
 */

typedef struct Execfile {
   ATHeader head;
   AssocTab *srcs;
   AssocTab *globals;
   AssocTab *gStructs;
   char *fileStart;
   pHDRR symHdr;      /* easy access */
   pFDR pfds;         /* easy access */
   pSYMR syms;        /* easy access */
   pAUXU aux;         /* easy access */
   pEXTR extrs;       /* easy access */
   pRFDT fit;         /* easy access */
} Execfile;

typedef struct Srcfile {
   ATHeader head;
   int num;
   FDR pfd;          /* easy access */
   pFDR pfdPtr;
} Srcfile;

typedef struct Global {
   ATHeader head;
   int isym;
   int filenum;
} Global;


/*
 *   variables
 */

static AssocTab *exectab = NULL;

static SymbInfo sitab[] = {
   {0, 0, 0, 0},   /* 0 tNil */
   {0, 1, 1, 1},   /* 1 tChar */
   {0, 1, 1, 1},   /* 2 tUChar */
   {0, 1, 1, 1},   /* 3 tShort */
   {0, 1, 1, 1},   /* 4 tUShort */
   {0, 1, 1, 1},   /* 5 tInt */
   {0, 1, 1, 1},   /* 6 tUInt */
   {0, 1, 1, 1},   /* 7 tLong */
   {0, 1, 1, 1},   /* 8 tULong */
   {0, 1, 1, 1},   /* 9 tFloat */
   {0, 1, 1, 1},   /* 10 tDouble */
   {0, 0, 1, 1},   /* 11 tStruct */
   {0, 0, 1, 1},   /* 12 tUnion */
   {0, 0, 1, 1},   /* 13 tEnum */
   {0, 0, 1, 1},   /* 14 tTypedef */
   {0, 0, 1, 1},   /* 15 tPtr */
   {0, 0, 1, 1},   /* 16 tArray */
   {0, 1, 0, 1},   /* 17 tLabel */
   {1, 0, 0, 1},   /* 18 tProc */
   {1, 1, 0, 0},   /* 19 tStructDef */
   {1, 1, 0, 0},   /* 20 tUnionDef */
   {1, 1, 0, 0},   /* 21 tEnumDef */
   {0, 1, 0, 0},   /* 22 tAddr */
   {1, 1, 0, 0},   /* 23 tFile */
   {0, 1, 0, 0},   /* 24 tEnd */
   {1, 1, 0, 0},   /* 25 tBlock */
   {0, 0, 0, 0},   /* 26 tTypedefDef */
   {0, 0, 0, 0},   /* 27 tIndirect */
};

SymbInfo *SymInfo = sitab;

/*
 *   functions
 */

static void FillIn(Symb *sym, Execfile *exec, int isym, int filenum);
static void FillInBasicType(Symb *sym, int index, pSYMR psym);

extern void *Simrmt_mmap(void *addr, size_t len, int prot, int flags,
                         char* pathname, off_t off);

static void CopySYMR (pSYMR  mapped,pSYMR  copy);
static void CopyFDR  (pFDR   mapped,pFDR   copy);
static void CopyTIR  (pTIR   mapped,pTIR   copy);
static void CopyRNDXR(pRNDXR mapped,pRNDXR copy);


/*
 *    SymLoad
 */

int SymLoad(char *pathname, char *execname)
{
   char *filename;
   struct stat statbuf;
   int fd, retval;
   char *fp;
   Execfile *exec;
   Srcfile *src;
   Global *global;
   ATHeader *dummy;
   FILHDR *fileHdr;
   Elf32_Ehdr *ehdr;
   pFDR pfd;
   pEXTR extr;
   int i;

   if (!exectab) {
      exectab = AssocTabCreate();
   }

   /* check if execname already used */
   if (AssocTabLookup(exectab, execname, &dummy) == AT_OK) {
      SymReportError("execname already taken \"", execname, "\"", NULL);
      return SYM_ERROR;
   }      

   if (!(filename = SearchForFile(pathname))) {
      SymReportError("no such file \"", pathname, "\"", NULL);
      return SYM_ERROR;
   }

   if (strchr(filename, ':')) {
      /* remote */
#if defined(SIMALPHA)
      ASSERT(0);
      fp = 0;
#else
      fp = Simrmt_mmap(NULL, 0, PROT_READ, MAP_SHARED, filename, 0);
#endif
   } else {
      /* local */
      fd = open(filename, O_RDONLY, 0);
      if (fd < 0) {
         SymReportError("couldn't open \"", filename, "\"", NULL);
         return SYM_ERROR;
      }
      
      retval = fstat(fd, &statbuf);
      if (retval < 0) {
         SymReportError("couldn't stat \"", filename, "\"", NULL);
         return SYM_ERROR;
      }
      
#ifdef __alpha
      fp = mmap(NULL, statbuf.st_size, PROT_READ, MAP_SHARED|MAP_FILE, fd, 0);
#else
      fp = mmap(NULL, statbuf.st_size, PROT_READ, MAP_SHARED, fd, 0);
#endif
      if (fp == (char*)-1) {
         SymReportError("couldn't mmap \"", filename, "\"", NULL);
         return SYM_ERROR;
      }

      close(fd);
   }
   
   fileHdr = (FILHDR*)fp;
   ehdr = (Elf32_Ehdr*)fp;

   exec = (Execfile*)malloc(sizeof(Execfile));
   exec->head.label = SaveString(execname);
   exec->fileStart = fp;
   
   if (ISCOFF(BE2HO_2(fileHdr->f_magic))) {
      if ((BE2HO_4(fileHdr->f_nsyms) == 0)
          || (BE2HO_4(fileHdr->f_symptr) == 0) 
          || (BE2HO_4(fileHdr->f_symptr) == -1)) {
         SymReportError("did not find symbol table in \"", filename, "\"", NULL);
         return SYM_ERROR;
      }
      
      exec->symHdr = (pHDRR)(fp + BE2HO_4(fileHdr->f_symptr));

   } else if (IS_ELF(*ehdr)) {
      Elf32_Shdr *shdr = (Elf32_Shdr *)(fp + BE2HO_4(ehdr->e_shoff));
      char *strs = (char *)fp + BE2HO_4(shdr[BE2HO_2(ehdr->e_shstrndx)].sh_offset);
      
      if ( BE2HO_1(ehdr->e_ident[EI_CLASS]) != ELFCLASS32 ) {
         /* should clean up mmapped stuff */
#ifdef DWARF2
         /* we add special entry with null fileStart to indicate dwarf */
         exec->fileStart = 0;
         if (AssocTabEnter(exectab, (ATHeader*)exec) != AT_OK) {
            ASSERT(0);
         }
         
         /* now call dwarf2 64-bit code */
         return Dwarf_SymLoad(pathname, execname);
#else
 	  /* don't fully support 64 bit elf files here yet */
         SymReportError("elf64 symtab format in \"", filename, "\"", NULL);
         return SYM_ERROR;
#endif
      }

      for (i=0; i<BE2HO_2(ehdr->e_shnum); i++, shdr++) {
         if (!strcmp(strs + BE2HO_4(shdr->sh_name), MIPS_MDEBUG)) {
            exec->symHdr = (pHDRR)(fp + BE2HO_4(shdr->sh_offset));
            break;
         }
      }

      if (i == ehdr->e_shnum) {
         SymReportError("did not find symbol table in \"", filename, "\"", NULL);
         return SYM_ERROR;
      }
         
   } else {
      SymReportError("architecture of \"", filename, "\" not known", NULL);
      return SYM_ERROR;
   }
   
   exec->pfds = (pFDR)  (fp + BE2HO_4(exec->symHdr->cbFdOffset));
   exec->syms = (pSYMR) (fp + BE2HO_4(exec->symHdr->cbSymOffset));
   exec->aux = (pAUXU)  (fp + BE2HO_4(exec->symHdr->cbAuxOffset));
   exec->extrs = (pEXTR)(fp + BE2HO_4(exec->symHdr->cbExtOffset));
   exec->fit = (pRFDT)  (fp + BE2HO_4(exec->symHdr->cbRfdOffset));
   
   exec->srcs = AssocTabCreate();
   exec->globals = AssocTabCreate();
   exec->gStructs = AssocTabCreate();

   pfd = (pFDR)(fp + BE2HO_4(exec->symHdr->cbFdOffset));
   for (i=0; i<BE2HO_4(exec->symHdr->ifdMax); i++, pfd++) {

      if (AssocTabLookup(exec->srcs,
                         StripPath(fp
                                   + BE2HO_4(pfd->issBase)
                                   + BE2HO_4(exec->symHdr->cbSsOffset)
                                   + BE2HO_4(pfd->rss)),
                         NULL) == AT_OK) {
         /* nothing */
      } else {
         src = (Srcfile*)malloc(sizeof(Srcfile));
         src->num = i;
         src->pfdPtr = pfd;
         CopyFDR(pfd,&src->pfd);
         src->head.label = StripPath(fp + src->pfd.issBase
                                     + BE2HO_4(exec->symHdr->cbSsOffset) 
                                     + src->pfd.rss);

         if (AssocTabEnter(exec->srcs, (ATHeader*)src) != AT_OK) {
            ASSERT(0);
         }

         if (src->pfd.fMerge) {
            /*
             * this is a header - lets scan it for structs.
             * even though structs are not global we are going to do this
             * as a heuristic
             */
            Symb symb;
            
            FillIn(&symb, exec, src->pfd.isymBase, i);

            if (SymIn(&symb) != SYM_OK) {
               ASSERT(0);
               return SYM_ERROR;
            }
            
            while (symb.type[symb.nextt] != tEnd) {
               
               if ((symb.type[symb.nextt] == tStructDef)
                   || (symb.type[symb.nextt] == tUnionDef)
                   || (symb.type[symb.nextt] == tEnumDef)) {

                  global = (Global*)malloc(sizeof(Global));
                  global->head.label = symb.name;
                  global->isym = symb.isym;
                  global->filenum = symb.filenum;
                  
                  if (AssocTabEnter(exec->gStructs, (ATHeader*)global) != AT_OK) {
                     /* 2 structs with the same name, ignore */
                  }
               }

               if (SymNext(&symb) != SYM_OK) {
                  ASSERT(0);
                  return SYM_ERROR;
               }
            }   
         }
      }
   }

   extr = exec->extrs;
   for (i=0; i<BE2HO_4(exec->symHdr->iextMax); i++, extr++) {
      global = (Global*)malloc(sizeof(Global));
      global->head.label = fp + BE2HO_4(exec->symHdr->cbSsExtOffset) + 
         BE2HO_4(extr->asym.iss);
      global->isym = i;
      global->filenum = -1;
      
      if (AssocTabEnter(exec->globals, (ATHeader*)global) != AT_OK) {
         /* 2 globals with the same name */
         ASSERT(0);
      }
   }

   if (AssocTabEnter(exectab, (ATHeader*)exec) != AT_OK) {
      ASSERT(0);
   }      

   return SYM_OK;
}


/*
 *   SymAddr2Symbol
 */

int SymAddr2Symbol(char *execname, int addr, Symb *sym)
{
   Execfile *exec = NULL;
   SYMR symr;
   pEXTR extr;
   int i;
   int best, bestFile, bestVal;
   int filenum;
   
   if (!exectab) {
      exectab = AssocTabCreate();
   }
   
   if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
      SymReportError("no exec named \"", execname, "\"", NULL);
      return SYM_ERROR;
   }

#ifdef DWARF2
   ASSERT (0);
#endif
#ifdef DWARF2
   if( exec->fileStart == 0 ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymAddr2Symbol(execname, addr, sym);
   }
#endif

   filenum = -1;
   bestVal = 0;
   best = -1;
   bestFile = -1;
   extr = exec->extrs;
   for (i=0; i<BE2HO_4(exec->symHdr->iextMax); i++, extr++) {
      if (((uint)BE2HO_4(extr->asym.value) > (uint)bestVal)
          && ((uint)BE2HO_4(extr->asym.value) <= (uint)addr)) {
         best = i;
         bestFile = filenum;
         bestVal =BE2HO_4(extr->asym.value);
      }
   }

   CopySYMR(exec->syms,&symr);
   for (i=0; i<BE2HO_4(exec->symHdr->isymMax); i++) {

      CopySYMR(exec->syms+i,&symr);

      if (symr.st == stFile) {
         filenum++;
      }

      if (symr.st == stLabel) {
         continue;
      }
      
      if ((symr.value > (uint)bestVal)
          && (symr.value <= (uint)addr)) {
         best = i;
         bestFile = filenum;
         bestVal = symr.value;
      }
   }

   FillIn(sym, exec, best, bestFile);
   return SYM_OK;
}


/*
 *   SymLine2Addr
 */

int SymLine2Addr(char *execname, char *srcfile, int line, Symb *sym)
{
   Execfile *exec = NULL;
   Srcfile *src = NULL;
   PDR *pdr;
   char *base;
   int bestaddr, p;
   
   if (!exectab) {
      exectab = AssocTabCreate();
   }
   
   if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
      SymReportError("no exec named \"", execname, "\"", NULL);
      return SYM_ERROR;
   }

#ifdef DWARF2
   ASSERT (0);
#endif
#ifdef DWARF2
   if( exec->fileStart == 0 ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymLine2Addr(execname, srcfile, line, sym);
   }
#endif

   if (AssocTabLookup(exec->srcs, StripPath(srcfile), (ATHeader**)&src) != AT_OK) {
      SymReportError("no source file named \"", srcfile, "\"", NULL);
      return SYM_ERROR;
   }
   
   base = exec->fileStart + 
      BE2HO_4(exec->symHdr->cbLineOffset) + 
      src->pfd.cbLineOffset;
   pdr = (pPDR)(exec->fileStart + BE2HO_4(exec->symHdr->cbPdOffset)) + 
      src->pfd.ipdFirst;

   bestaddr = src->pfd.adr;
   
   for (p=0; p<src->pfd.cpd; p++, pdr++) {
      char *ptr, *end;
      int curline, curaddr, delta, num;

      if (line < BE2HO_4(pdr->lnLow)) {
         bestaddr = BE2HO_4(pdr->adr) + src->pfd.adr;
         break;
      } else if (line > BE2HO_4(pdr->lnHigh)) {
         if ((p+1) < src->pfd.cpd) {
            continue;
         } else {
            /* what do I do here? */
            line = BE2HO_4(pdr->lnHigh);
         }
      }
      
      curline = BE2HO_4(pdr->lnLow);
      curaddr = BE2HO_4(pdr->adr) +src->pfd.adr;
      
      ptr = base + BE2HO_4(pdr->cbLineOffset);

      if ((p+1) < src->pfd.cpd) {
         end = base + BE2HO_4(pdr[1].cbLineOffset);
      } else {
         end = base + src->pfd.cbLine;
      }
      
      while (ptr < end) {
         num = (*ptr & 0x0F) + 1;
         delta = *ptr >> 4;
         
         ptr++;
         
         if (delta >= 8) {
            delta -= 16;
         }
         
         if (delta == -8) {
            delta = (ptr[0] << 8) | ptr[1];
            if (delta >= 0x8000) {
               delta -= 0x10000;
            }
            ptr += 2;
         }
         
         curline += delta;

         if (curline < line) {
            bestaddr = curaddr + 4 * num;
         }

         curaddr += 4 * num;
      }

      break;
   }

   sym->global = 0;
   sym->isym = -1;
   sym->filenum = src->num;
   sym->name = NULL;
   sym->index = -1;
   sym->exec = exec;
   
   sym->type[1] = tAddr;
   sym->type[2] = tNil;
   sym->nextt = 1;
   
   sym->value = bestaddr;
   
   return SYM_OK;
}


/*
 *   SymTypeCast
 */

int SymTypeCast(char *execname, char *srcfile, int *types, char *name, Symb *sym)
{
   Symb tmp;
   int *tp;
   
#ifdef DWARF2
   Execfile *exec = NULL;

   ASSERT (exectab);

   if (!exectab) {
      exectab = AssocTabCreate();
   }

   if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
      SymReportError("no exec named \"", execname, "\"", NULL);
      return SYM_ERROR;
   }

   ASSERT (exec->fileStart==0);
   if( exec->fileStart == 0 ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymTypeCast(execname, srcfile, types, name, sym);
   }
#endif

   sym->global = 0;
   sym->isym = -1;
   sym->filenum = -1;
   sym->name = NULL;
   sym->index = -1;
   sym->exec = NULL;

   sym->nextt = 1;
   tp = types;

   while (*tp != tNil) {
      sym->type[sym->nextt++] = *tp;
      tp++;
   }
   
   ASSERT(sym->nextt<MAX_TYPES);
   sym->type[sym->nextt] = tNil;

   if (name[0]) {

      sym->nextt--; /* OK because we are going to reset it anyway */

      if ((sym->type[sym->nextt] == tStruct)
          || (sym->type[sym->nextt] == tUnion)
          || (sym->type[sym->nextt] == tEnum)) {
         if (SymLookup(execname, srcfile, name, SYM_TAGS, &tmp) != SYM_OK) {
            return SYM_ERROR;
         }

      } else if (sym->type[sym->nextt] == tTypedef) {
         if (SymLookup(execname, srcfile, name, 0, &tmp) != SYM_OK) {
            return SYM_ERROR;
         }

      } else {
         ASSERT(0);
      }
      
      sym->exec = tmp.exec;
      sym->filenum = tmp.filenum;
      sym->rndx.rfd = tmp.filenum;
      sym->rndx.index = tmp.isym - BE2HO_4(tmp.exec->pfds[tmp.filenum].isymBase);
   }
   
   sym->nextt = 1;
   sym->value = 0;
   
   return SYM_OK;
}


/*
 *   FillIn
 *
 *     private function which fills in the fields of a symbol
 */

void FillIn(Symb *sym, Execfile *exec, int isym, int filenum)
{
   pFDR pfdPtr;
   FDR  frd;
   SYMR symr;
   uint index;
   if (filenum == -1) {
      sym->global = 1;
      sym->isym = isym;
      sym->filenum = BE2HO_2(exec->extrs[isym].ifd);
      sym->name = exec->fileStart + 
         BE2HO_4(exec->symHdr->cbSsExtOffset)
         + BE2HO_4(exec->extrs[isym].asym.iss);
      CopySYMR(&exec->extrs[isym].asym,&symr);
      
   } else {
      sym->global = 0;
      sym->isym = isym;
      sym->filenum = filenum;
      sym->name = exec->fileStart + 
         BE2HO_4(exec->symHdr->cbSsOffset)
         + BE2HO_4(exec->pfds[filenum].issBase) 
         + BE2HO_4(exec->syms[isym].iss);
      CopySYMR(&exec->syms[isym],&symr);
   }
   ASSERT( isym != 918659);


   pfdPtr = &(exec->pfds[sym->filenum]);
   CopyFDR(pfdPtr,&frd);

#ifdef notdef
   SIM_DEBUG(('t', "filling in symbol %s file %d isym %d global %d\n\r",
              sym->name, sym->filenum, sym->isym, sym->global));
   SIM_DEBUG(('t', "     file %s auxs %d\n\r", exec->fileStart
              + frd.issBase + exec->symHdr->cbSsOffset + frd.rss,
              frd.iauxBase));
#endif
   
   sym->exec = exec;
   sym->nextt = 1;
   sym->index = symr.index;
   ASSERT( sym->index != 0x18200000);

   index = sym->index  + frd.iauxBase;
   switch (symr.st) {
   case stProc:
   case stStaticProc: {
      sym->type[sym->nextt++] = tProc;
      if (sym->global) {
         pSYMR symrPtr = &exec->syms[symr.index + frd.isymBase];
         SYMR symr2;
         CopySYMR(symrPtr,&symr2);
         sym->index = symr2.index;
         ASSERT(  sym->index != 0x18200000);
      }
      if (sym->index != indexNil) {
         FillInBasicType(sym,sym->index + frd.iauxBase + 1, &symr);
      }
      break;
   }
   case stGlobal:
   case stStatic:
   case stMember: {
      if (sym->index != indexNil) {
         FillInBasicType(sym, index, &symr);
      }
      break;
   }
   case stTypedef: {
      sym->type[sym->nextt++] = tTypedefDef;
      if (sym->index != indexNil) {
         FillInBasicType(sym, index, &symr);
      }
      break;
   }
   case stIndirect: {
      pRNDXR x = &sym->exec->aux[index + 1].rndx;
      CopyRNDXR(x,&sym->rndx);
      sym->type[sym->nextt++] = tIndirect;
      sym->fip = BE2HO_4(sym->exec->aux[index + 2].isym);
   }
   case stLabel: {
      sym->type[sym->nextt++] = tLabel;
      break;
   }
   case stStruct: {
      sym->type[sym->nextt++] = tStructDef;
      break;
   }
   case stUnion: {
      sym->type[sym->nextt++] = tUnionDef;
      break;
   }
   case stEnum: {
      sym->type[sym->nextt++] = tEnumDef;
      break;
   }
   case stFile: {
      sym->type[sym->nextt++] = tFile;
      break;
   }
   case stEnd: {
      sym->type[sym->nextt++] = tEnd;
      break;
   }
   case stBlock: {
      sym->type[sym->nextt++] = tBlock;
      break;
   }
   case stRegReloc:
   case stParam:
   case stLocal:
   case stNil:
   default:
      break;
   }

   ASSERT(sym->nextt<MAX_TYPES);
   sym->type[sym->nextt] = tNil;
   sym->nextt = 1;
   sym->value = symr.value;
}

/* *********************************************
 * psym is a copy passed in the host order!!!
 * *********************************************/

void FillInBasicType(Symb *sym, int index, pSYMR psym)
{
   pTIR tirPtr;
   TIR tirCpy;

   int tq, i;
   uint sc =psym->sc;
   
   if (index == -1) {
      sym->type[sym->nextt++] = tNil;
      return;
   }
   
   if (sc == scAbs) {
      SIM_DEBUG(('t', "skipping symbol %s st %d sc %d\n\r", sym->name, psym->st, 
                 psym->sc));
      return;
   }

   if ((sc != scNil) && (sc != scText)
       && (sc != scBss) && (sc != scSBss)
       && (sc != scSData) && (sc != scInfo)
       && (sc != scData)) {
      printf("symbol %s st %d sc %d\n\r", sym->name, psym->st, sc);
      ASSERT(0);
   }
   
   tirPtr = &(sym->exec->aux[index].ti);
   CopyTIR(tirPtr,&tirCpy);
   ASSERT(!tirCpy.continued);
   ASSERT(!tirCpy.fBitfield);

   for (i=5; i>=0; i--) {
      switch (i) {
      case 0:
         tq = tirCpy.tq0;
         break;
      case 1:
         tq = tirCpy.tq1;
         break;
      case 2:
         tq = tirCpy.tq2;
         break;
      case 3:
         tq = tirCpy.tq3;
         break;
      case 4:
         tq = tirCpy.tq4;
         break;
      case 5:
         tq = tirCpy.tq5;
         break;
      default:
         ASSERT(0);
         tq = -1;
      }

      if (tq == tqNil) {
      } else if (tq == tqPtr) {
         sym->type[sym->nextt++] = tPtr;
      } else if (tq == tqArray) {
         sym->type[sym->nextt++] = tArray;
      } else if (tq == tqProc) {
         sym->type[sym->nextt++] = tPtr;
      } else if ((tq == tqVol) || (tq == tqConst)) {
         /* ignore */
      } else {
         ASSERT(0);
      }
   }

   if ((tirCpy.bt == btStruct)
       || (tirCpy.bt == btUnion)
       || (tirCpy.bt == btEnum)
       || (tirCpy.bt == btTypedef)) {
      pRNDXR x =&sym->exec->aux[index+1].rndx; 
      CopyRNDXR(x,&sym->rndx);
      sym->fip = BE2HO_4(sym->exec->aux[index+2].isym);
   }
   
   if ((tirCpy.bt >= btChar) && (tirCpy.bt <= btTypedef)) {
      sym->type[sym->nextt++] = (tirCpy.bt - btChar) + tChar;
   } else {
      sym->type[sym->nextt++] = tNil;
   }
}


/*
 *   SymLookup
 */

int SymLookup(char *execname, char *srcfile, char *symname, int flags, Symb *sym)
{
   Execfile *exec = NULL;
   Srcfile *src = NULL;
   Global *global = NULL;
   pFDR pfdPtr;
   FDR frd;
   int filenum;
   
   if (!exectab) {
      exectab = AssocTabCreate();
   }

   if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
      SymReportError("no exec named \"", execname, "\"", NULL);
      return SYM_ERROR;
   }

#ifdef DWARF2
   ASSERT (0);
#endif
#ifdef DWARF2
   if( exec->fileStart == 0 ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymLookup(execname, srcfile, symname, flags, sym);
   }
#endif

   if (srcfile[0]) {
      if (AssocTabLookup(exec->srcs, StripPath(srcfile),
                         (ATHeader**)&src) != AT_OK) {
         SymReportError("no source file named \"", srcfile, "\"", NULL);
         return SYM_ERROR;
      }

      filenum = src->num;
      pfdPtr = src->pfdPtr;
      CopyFDR(pfdPtr,&frd);
      do {
         FillIn(sym, exec, frd.isymBase, filenum);

         if (frd.lang == langAssembler) {
            /* the structure of assembly files is all screwed up */
            flags |= SYM_RECURSE;
         }
         
         if (SymFind(sym, symname, flags) == SYM_OK) {
            return SYM_OK;
         }
         
         pfdPtr++;
         CopyFDR(pfdPtr,&frd);
         filenum++;
         
      } while (frd.fMerge);
   }

   /* search globals */
   if (flags & SYM_TAGS) {
      if (AssocTabLookup(exec->gStructs, symname, (ATHeader**)&global) != AT_OK) {
         SymReportError("couldn't find symbol \"", symname, "\"", NULL);
         return SYM_ERROR;
      }
      
      FillIn(sym, exec, global->isym, global->filenum);

   } else {
      if (AssocTabLookup(exec->globals, symname, (ATHeader**)&global) != AT_OK) {
         SymReportError("couldn't find symbol \"", symname, "\"", NULL);
         return SYM_ERROR;
      }
      
      FillIn(sym, exec, global->isym, -1);
   }
   
   return SYM_OK;
}


/*
 *   SymFind
 */

int SymFind(Symb *sym, char *name, int flags)
{
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymFind(sym, name, flags);
   }
#endif

   if (SymIn(sym) != SYM_OK) {
      ASSERT(0);
      return SYM_ERROR;
   }
   
   while (sym->type[sym->nextt] != tEnd) {

      if (!strcmp(sym->name, name)) {

         if ((sym->type[sym->nextt] == tStructDef)
             || (sym->type[sym->nextt] == tUnionDef)
             || (sym->type[sym->nextt] == tEnumDef)) {

            if (flags & SYM_TAGS) {
               return SYM_OK;
            }
            
         } else if (!(flags & SYM_TAGS)) {
            return SYM_OK;
         }
      }

      /* we effectively ignore blocks */
      if ((sym->type[sym->nextt] == tBlock)
          || ((flags & SYM_RECURSE) && (sym->type[sym->nextt] == tProc))) {
         Symb in = *sym;

         if (SymFind(&in, name, flags) == SYM_OK) {
            *sym = in;
            return SYM_OK;
         }
      }
      
      if (SymNext(sym) != SYM_OK) {
         ASSERT(0);
         return SYM_ERROR;
      }
   }   

   SymReportError("couldn't find symbol \"", name, "\"", NULL);
   return SYM_ERROR;
}


/*
 *   SymNext
 */

int SymNext(Symb *sym)
{
   Execfile *exec = sym->exec;
   pFDR pfdPtr;
   FDR fdr;
   int isym = sym->isym;
   
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymNext(sym);
   }
#endif

   pfdPtr = &(exec->pfds[sym->filenum]);
   CopyFDR(pfdPtr,&fdr);
   if (sym->type[sym->nextt] == tEnd) {
      SymReportError("internal error, next over end", NULL);
      return SYM_ERROR;
   }

   if (sym->global) {
      isym++;
      
   } else if (sym->type[sym->nextt] == tFile) {
      ASSERT(0);
      
   } else if (sym->type[sym->nextt] == tProc) {
      isym = fdr.isymBase + BE2HO_4(exec->aux[sym->index + fdr.iauxBase].isym);
      
   } else if ((sym->type[sym->nextt] == tStructDef)
              || (sym->type[sym->nextt] == tUnionDef)
              || (sym->type[sym->nextt] == tEnumDef)
              || (sym->type[sym->nextt] == tBlock)) {
      isym = fdr.isymBase + sym->index;
      
   } else {
      isym++;
   }

   FillIn(sym, exec, isym, (sym->global) ? -1 : sym->filenum);
   return SYM_OK;
}

/*
 *   SymIn
 */

int SymIn(Symb *sym)
{
   Execfile *exec = sym->exec;
   pFDR pfdPtr;
   int isym = sym->isym;

#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymIn(sym);
   }
#endif

   pfdPtr = &(exec->pfds[sym->filenum]);
   
   if (!SymIsHier(sym)) {
      SymReportError("internal error, in into non-hier", NULL);
      return SYM_ERROR;
   }

   if ((sym->type[sym->nextt] == tProc) && (sym->global)) {
      pSYMR symrPtr = &exec->extrs[sym->isym].asym;
      SYMR symr;
      CopySYMR(symrPtr,&symr);
      isym = BE2HO_4(pfdPtr->isymBase) + symr.index;
   } else {
      ASSERT(!sym->global);
   }
   
   isym++;

   FillIn(sym, exec, isym, sym->filenum);
   return SYM_OK;
}


/*
 *   SymSubType
 */

int SymSubtype(Symb *sym, int flags)
{
   Execfile *exec = sym->exec;
   int filenum;
   int st;
   int code;
   
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymSubtype(sym, flags);
   }
#endif

  if (SymIsBasicType(sym)) {
      SymReportError("internal error, subtype of basic", NULL);
      return SYM_ERROR;

   } else if ((sym->type[sym->nextt] == tPtr) || (sym->type[sym->nextt] == tArray)) {
      sym->nextt++;
      ASSERT(SymIsVarType(sym));
      
   } else  {

      st = sym->type[sym->nextt];
      
      ASSERT((st == tStruct) || (st == tUnion) || (st == tEnum)
             || (st == tTypedef) || (st == tIndirect));
   
      if (sym->rndx.rfd == 4095) {
         int x = BE2HO_4(exec->pfds[sym->filenum].rfdBase);
         filenum = BE2HO_4(exec->fit[x + sym->fip]);
      } else {
         filenum = sym->rndx.rfd;
      }
      
      FillIn(sym, exec, sym->rndx.index 
             + BE2HO_4(exec->pfds[filenum].isymBase), filenum);

      /* ugliness to skip indirects */
      if (sym->type[sym->nextt] == tIndirect) {
         if (sym->rndx.index == 0) {
            SymReportError("definition not found in source file", NULL);
            return SYM_ERROR;
         }
         code = SymSubtype(sym, 0);
         ASSERT(code == SYM_OK);
      }
         
      if (st == tStruct) {
         ASSERT(sym->type[sym->nextt] == tStructDef);
         
      } else if (st == tUnion) {
         ASSERT(sym->type[sym->nextt] == tUnionDef);
         
      } else if (st == tEnum) {
         ASSERT(sym->type[sym->nextt] == tEnumDef);
         
      } else if (st == tTypedef) {
         ASSERT(sym->type[sym->nextt] == tTypedefDef);
         sym->nextt++;
         ASSERT(SymIsVarType(sym));

         if (flags & SYM_RESOLVE) {
            return SymSubtype(sym, flags);
         }

      } else if (st == tIndirect) {
         ASSERT((sym->type[sym->nextt] == tStructDef)
                || (sym->type[sym->nextt] == tUnionDef));

      } else {
         ASSERT(0);
      }
   }

   return SYM_OK;
}


/*
 *   SymType
 */

int SymType(Symb *sym, int flags)
{
   Symb tmp;
   
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
      /* hack to indicate must be dwarf stuff */
      return Dwarf_SymType(sym, flags);
   }
#endif

   if ((sym->type[(sym)->nextt] != tTypedef) || !(flags & SYM_RESOLVE)) {
      return sym->type[(sym)->nextt];
   }

   tmp = *sym;
   SymSubtype(&tmp, 0);
   return SymType(&tmp, flags);
}


/*
 *   SymSize
 */

Reg_s SymSize(Symb *sym)
{
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymSize(sym);
   }
#endif

   switch (SymType(sym, SYM_RESOLVE)) {
   case tChar:
   case tUChar:
      return 1;
   case tShort:
   case tUShort:
      return 2;
   case tInt:
   case tUInt:
   case tLong:
   case tULong:
   case tPtr:
      return 4;
   case tFloat:
      return sizeof(float);
   case tDouble:
      return sizeof(double);
   case tLabel:
   case tAddr:
      return 4;
   case tProc: {
      pFDR pfd = &sym->exec->pfds[sym->filenum];
      int isym = BE2HO_4(sym->exec->aux[sym->index + BE2HO_4(pfd->iauxBase)].isym) + 
         BE2HO_4(pfd->isymBase) - 1;
      return BE2HO_4(sym->exec->syms[isym].value);
   }
   case tArray:
      /* XXX fix */
      return 4;
   case tStruct:
   case tUnion:
   case tEnum: {
      Symb tmp = *sym;
      if (SymSubtype(&tmp, SYM_RESOLVE) != SYM_OK) {
         ASSERT(0);
      }
      return tmp.value;
   }
   case tStructDef:
   case tUnionDef:
   case tEnumDef:
      return sym->value;
   default:
      ASSERT(0);
      break;
   }

   /* never reached - for compiler */
   ASSERT(0);
   return 0;
}


/*
 *   SymMakeAddr
 *
 *     This is not an rpc because we have all the info
 *     we need locally.
 */

int SymMakeAddr(Symb *sym)
{
#ifdef DWARF2
   ASSERT (sym->index == DWARF_INDEX);
   if( sym->index == DWARF_INDEX ) {
       /* hack to indicate must be dwarf stuff */
       return Dwarf_SymMakeAddr(sym);
   }
#endif
   ASSERT(sym->nextt > 0);
   sym->nextt--;
   sym->type[sym->nextt] = tAddr;

   return SYM_OK;
}


static void CopySYMR(pSYMR mapped,pSYMR copy)
{
   int i;
   char * mptr = (char*)mapped;
   char * cptr = (char*)copy;
   for (i=0;i<sizeof(SYMR);i+=4) {
      uint *mp = (uint*)(mptr+i);
      uint *cp = (uint*)(cptr+i);
      *cp = BE2HO_4(*mp);
   }
}

static void CopyFDR(pFDR mapped,pFDR copy)
{
   int i;
   char * mptr = (char*)mapped;
   char * cptr = (char*)copy;
   for (i=0;i<sizeof(FDR);i+=4) {
      uint *mp = (uint*)(mptr+i);
      uint *cp = (uint*)(cptr+i);
      *cp = BE2HO_4(*mp);
   }
   copy->ipdFirst = BE2HO_2(mapped->ipdFirst);
   copy->cpd      = BE2HO_2(mapped->cpd);
}

static void CopyTIR(pTIR mapped,pTIR copy)
{
   int i;
   char * mptr = (char*)mapped;
   char * cptr = (char*)copy;
   for (i=0;i<sizeof(TIR);i+=4) {
      uint *mp = (uint*)(mptr+i);
      uint *cp = (uint*)(cptr+i);
      *cp = BE2HO_4(*mp);
   }
}


static void CopyRNDXR(pRNDXR mapped, pRNDXR copy)
{
   int i;
   char * mptr = (char*)mapped;
   char * cptr = (char*)copy;
   for (i=0;i<sizeof(RNDXR);i+=4) {
      uint *mp = (uint*)(mptr+i);
      uint *cp = (uint*)(cptr+i);
      *cp = BE2HO_4(*mp);
   }
}

