/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <filehdr.h>
#include <stddef.h>

#include "simtypes.h"
#include "sim_error.h"
#include "alpha.h"
#include "a_out.h"
#include "new_aouthdr.h"
#include "alpha_access.h"
#include "addr_layout.h"

static int debugLoader = 1;

struct aouthdr {
	short	magic;		/* see above				*/
	short	vstamp;		/* version stamp			*/
	long	tsize;		/* text size in bytes, padded to DW bdry*/
	long	dsize;		/* initialized data "  "		*/
	long	bsize;		/* uninitialized data "   "		*/
	long	entry;		/* entry pt.				*/
	long	text_start;	/* base of text used for this file	*/
	long	data_start;	/* base of data used for this file	*/
	long	bss_start;	/* base of bss used for this file	*/
	long	gprmask;	/* general purpose register mask	*/
	long	cprmask[4];	/* co-processor register masks		*/
	long	gp_value;	/* the gp value used for this object    */
};

#if 0 
#define AOUTHDR struct aouthdr
#define AOUTHSZ sizeof(AOUTHDR)
#define	SCNHDR	struct scnhdr
#define	SCNHSZ	sizeof(SCNHDR)
#define	FILHDR	struct filehdr
#define	FILHSZ	sizeof(FILHDR)
#endif

#define	 LIBMAGIC	0443


/*
 * mainly ok to load pal code, not regular a.out format
 */

int load_ultrix_aout (char *aout_file, VA address)
{
   int      fd;
   struct   exec     exec;
   struct   filehdr  file_header;
   int i;
   VA vA = ALIGN4(address);
  
   if((fd = open (aout_file, 0)) == -1)   {
      CPUError( "Cannot open %s \n", aout_file);
   };
   
   /* Read the file exec information */
   i = read(fd,&file_header,sizeof(struct filehdr));
   lseek(fd,0,SEEK_SET);
   i = read (fd, &exec, sizeof (struct exec));
   
   /*
    * Only support omagic files
    */
   ASSERT(i==sizeof(struct exec));
   if (exec.a_magic!=OMAGIC && exec.a_magic!=NMAGIC) {
      close(fd);
      return 0;
   }

   lseek (fd, N_TXTOFF (exec), SEEK_SET);
   if( debugLoader ) {
      CPUWarning ("Loading %s text at vAddr %lx memAddr %lx len=%x\n", 
		  aout_file, vA, PHYS_TO_MEMADDR(0,vA),exec.a_text);
   }
   i = read(fd,PHYS_TO_MEMADDR(0,vA),exec.a_text);
   if (i!=exec.a_text) { 
      perror("load_ultrix_aout. ");
      ASSERT(0);
   }
   if(debugLoader) CPUWarning("entry at %x\n", exec.a_entry);
   return TRUE;
}


/*
 * loading the kernel. assume kseg addresses --> phys by dropping hi
 */

int load_ultrix_coff(char *filename)
{
   int fd,i;
   struct new_aouthdr new_exec;	/* new optional header	*/
   struct new_filehdr new_fhdr;	/* new file header	*/
 
   if((fd = open (filename, 0)) == -1)   {
      CPUError( "Cannot open %s \n", filename);
   };
   i = read (fd, &new_fhdr.f_magic, sizeof (new_fhdr.f_magic));
   if (i != sizeof (new_fhdr.f_magic)) {
      close (fd);
      ASSERT(0);
   }
   if (new_fhdr.f_magic !=ALPHAMAGIC) { 
      return 0;
   }

   i = read (fd, (char *)&new_fhdr + sizeof(new_fhdr.f_magic),
	     sizeof(new_fhdr) - sizeof(new_fhdr.f_magic));
   ASSERT(i == sizeof(new_fhdr) - sizeof(new_fhdr.f_magic));

   i = read (fd, &new_exec, sizeof (new_exec));
   ASSERT(i==sizeof(new_exec));
   ASSERT( new_exec.magic == OMAGIC  ||  new_exec.magic == NMAGIC);

   lseek (fd, NEW_C_TXTOFF(new_fhdr,new_exec), 0);
    
   CPUWarning("OS start at phys addr %lx \n",new_exec.text_start);
   
   /*
    * start pc (entry point)
    */
   i = read(fd,PHYS_TO_MEMADDR(0,new_exec.text_start),
	    new_exec.tsize);
   ASSERT(i==new_exec.tsize);
   i = read(fd,PHYS_TO_MEMADDR(0,new_exec.data_start),
	    new_exec.dsize);
   ASSERT(i==new_exec.dsize);
   ASSERT(new_exec.text_start+new_exec.tsize==new_exec.data_start);

   /*
    * copy arguments of file header 
    */
   lseek(fd,0,0);
   i = NEW_FILHSZ+NEW_AOUTHSZ+(new_fhdr.f_nscns*NEW_SCNHSZ);
   ASSERT( i <1024);
#ifdef gone
   read(fd,alphaAccess.kernelFileHdr,i);
#endif

   /*
    * XXX bugnion
    * Major hack. Make sure that the kernel is loaded after the console
    * or we are in trouble
    */
   
   alphaAccess.kernStart = 
      ((uint64)new_exec.text_start_hi<<32) +  new_exec.text_start;

   alphaAccess.kernEnd =  
      ((uint64)new_exec.bss_start_hi<<32) + new_exec.bss_start + new_exec.bsize;

   alphaAccess.entryPoint = ((uint64)new_exec.entry_hi<<32)+  new_exec.entry;

   alphaAccess.last_offset = sizeof(struct AlphaAccess);

   CPUWarning("Loading COFF, nsec=%d start=0x%lx end=0x%lx entry=0x%lx \n",
	      new_fhdr.f_nscns,
	      alphaAccess.kernStart,
	      alphaAccess.kernEnd,
	      alphaAccess.entryPoint);
   return 1;

}


#define MAX_NT_PALCODE (64*1024)
/*
 * In this file format, the first 0x200 bytes seem to be garbage at
 * best. We load them anyway into memory. The exception vectors
 * start for real at +0x200
 */

void LoadNT_PALcode(char *filename, VA address)
{
   int fd,i;
   uint header;
   ASSERT(address >= 0x200);
   if((fd = open (filename, 0)) == -1)   {
      CPUError( "Cannot open %s \n", filename);
   }
   i = read (fd, &header,sizeof(header));

   lseek(fd,0,0);
   i = read(fd,PHYS_TO_MEMADDR(0,address-0x200),MAX_NT_PALCODE);
   if (i==MAX_NT_PALCODE) { 
      CPUError("PALcode too large. Increase loader.c::MAX_NT_PALCODE\n");
   }
   
}





