/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "tlaser.h"
#include "devices.h"


static struct TlaserUARTState {
   int statusStore;
   int nextChar;
} tlaserUartState[SIM_MAXCPUS];



int ALPHA_8530_CN_access(int cpuNum, VA vA, int type, Reg *buff)
{
   VA ALPHA_8530_CN_BASE = GBUS_BASE + UART_BASE;
   VA pc = PE[cpuNum]->PC;
   VA ra;

   ASSERT(ALPHA_8530_CN_BASE == 0xfffffcffa0000000);
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

#if 0
   CPUPrint(" ALPHA_8530 :  PC %lx RA 0x%lx va=0x%lx type=%s \n",
            pc,
            ra,
            vA,bdoorType[type]);	 
#endif   
   /*
    * Status Register 
    */
      
   if (vA == (ALPHA_8530_CN_BASE + 0x80) && BDOOR_IS_LOAD(type)) {
      int status = sim_console_int_status_noSE(0);
      if (tlaserUartState[cpuNum].nextChar < 0) {
         tlaserUartState[cpuNum].nextChar = sim_console_in(0);
         if (tlaserUartState[cpuNum].nextChar < 0) {
            status &= ~CONS_INT_RX;
         }
      } else {
         status |= CONS_INT_RX;
      }
     
      if (tlaserUartState[cpuNum].statusStore == 3) {
         /*
          * RR3 stuff? Don't really understand it, btw 
          */
         tlaserUartState[cpuNum].statusStore = 0;
         if (status & CONS_INT_TX) {
            *buff = (1<<4);
            return 0;
         } else if (status & CONS_INT_RX) {
            *buff = 1<<5;
            return 0;
         } else {
            CPUError("ALPHA_8530_CN_access: spurious access \n");
         }
         

      } else { 
         int reg =(1<<2);
         if (status & CONS_INT_RX) {
            reg |= (1<<0);
         }
         *buff = reg;
         return 0;

      }
   }
   if (vA == (ALPHA_8530_CN_BASE + 0x80) && BDOOR_IS_STORE(type)){
      tlaserUartState[cpuNum].statusStore = *buff;
      if (*buff==3) {
         return 0; /* going to read RR3 */
      }
      if (*buff==0x28) {
         /* Ack of TX */
         int status = sim_console_int_status(0);
         ASSERT( status & CONS_INT_TX);
         return 0;
      }
      if (*buff==0x1 || *buff==0 || *buff==0x12 ) {
         /* going to write data??? */
         return 0;
      }
      CPUWarning("ALPHA_8530_CN_access. writing status register 0x%lx \n",*buff);
      return 0;
   }

   /*
    * Data register. RX on read, TX on write
    */
   if (vA == (ALPHA_8530_CN_BASE + 0xc0) && BDOOR_IS_STORE(type)) {
      sim_console_out(0,*(Reg*)buff);
      if (pc != PE[cpuNum]->PC)
	CPUWarning("ALPHA_8530_CN_access: oldpc=%lx newpc=%lx\n",
		   pc, PE[cpuNum]->PC);
      return (pc == PE[cpuNum]->PC ?MMU_SUCCESS:MMU_EXCEPTION);
   }
   
   if (vA == (ALPHA_8530_CN_BASE + 0xc0) && BDOOR_IS_LOAD(type)) {
      ASSERT(tlaserUartState[cpuNum].nextChar >=0);
      *buff = tlaserUartState[cpuNum].nextChar;
      tlaserUartState[cpuNum].nextChar = -1;
      return 0;
   }
   
   CPUError(" ALPHA_8530 :  PC %lx RA 0x%lx va=0x%lx type=%s *buff=%lx\n",
	    PE[cpuNum]->PC,
	    ra,
	    vA,bdoorType[type], *buff);	 
   switch(type) { 
   case BDOOR_LOAD_WORD:
   case BDOOR_LOAD_DOUBLE:
   case BDOOR_LOAD_HALF:
   case BDOOR_LOAD_BYTE:
      *(uint64*)buff = 0;
      break;
   default: 
      break;
   }
   return SUCCESS;
}


void InitTLaserUart(void)
{
   int i;
   for (i=0;i<SIM_MAXCPUS;i++) {
      tlaserUartState[i].nextChar = -1;
   }
}
