/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * simfirewall.h
 *
 * Firewall implementation.
 * Dan Teodosiu, 07/96
 * John Chapin, 08/96
 *
 ****************************************************************/

#ifndef _SIM_FIREWALL_H
#define _SIM_FIREWALL_H

extern void sim_firewall_init(int restoreFromCkpt, int n, int pg_per_node);

/* get page protection for specified page. Bits as defined above */
extern uint64 sim_firewall_page_prot(int n, int pg);

/* set page protection for page pg of node n.
 * cpuNum is the cpu attempting to set the firewall.  Returns
 * 0 if successful, 1 if cpu write attempt is denied.
 */
extern int sim_firewall_set_page_prot(int cpuNum, int node, 
                                      int pg, uint64 prot);

/* set page protection for page pg of node n.
 * Directly modifies data structure; doesn't check legality of modification
 * or notify currently executing CPU simulator of the change.
 */
extern void sim_firewall_set_page_prot_raw(int node, int pg, uint64 prot);


#define SIMFIREWALL_ON (CPUVec.CheckFirewall != 0)

/* Note: CheckFirewall has now moved to CPUVec. It used to return
   1 if access allowed (return undefined if SIMFIREWALL_ON == 0). */

#endif /* _SIM_FIREWALL_H */
